r"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\
 
Copyright 2024 Zahner-Elektrik GmbH & Co. KG
 
Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:
 
The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.
 
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

from enum import Enum, IntEnum
import re
import shutil
import os
from typing import Optional, Union, Any, List

from thales_remote.error import ThalesRemoteError, TermConnectionError
from thales_remote.connection import ThalesRemoteConnection

MINIMUM_THALES_VERSION = "5.9.3"


def versiontuple(v):
    return tuple(map(int, (v.split("."))))


class PotentiostatMode(IntEnum):
    r"""
    Working modes for the potentiostat
    """

    POTMODE_POTENTIOSTATIC = 1
    POTMODE_GALVANOSTATIC = 2
    POTMODE_PSEUDOGALVANOSTATIC = 3


class ScanStrategy(IntEnum):
    r"""
    Options for the EIS scan strategy

    * SINGLE_SINE: single frequency sweep
    * MULTI_SINE: multi sine
    * TABLE: frequency table
    """

    SINGLE_SINE = 0
    MULTI_SINE = 1
    TABLE = 2

    @classmethod
    def stringToEnum(cls, string: str):
        stringEnumMap = {
            "single": ScanStrategy.SINGLE_SINE,
            "multi": ScanStrategy.MULTI_SINE,
            "table": ScanStrategy.TABLE,
        }
        if not string in stringEnumMap:
            raise ValueError("invalid string: " + string)
        return stringEnumMap.get(string)


class ScanDirection(IntEnum):
    r"""
    Set the scan direction for EIS measurements.

    * START_TO_MAX: from the start frequency to the maximum frequency
    * START_TO_MIN: from the start to the minimum frequency
    """

    START_TO_MAX = 0
    START_TO_MIN = 1

    @classmethod
    def stringToEnum(cls, string: str):
        stringEnumMap = {
            "startToMax": ScanDirection.START_TO_MAX,
            "startToMin": ScanDirection.START_TO_MIN,
        }
        if not string in stringEnumMap:
            raise ValueError("invalid string: " + string)
        return stringEnumMap.get(string)


class FileNaming(IntEnum):
    r"""
    Options for the file names in Thales.

    * DATE_TIME: naming with time stamp
    * INDIVIDUAL: only the specified filename without extension
    * COUNTER: consecutive number
    """

    DATE_TIME = 0
    COUNTER = 1
    INDIVIDUAL = 2

    @classmethod
    def stringToEnum(cls, string: str):
        stringEnumMap = {
            "dateTime": FileNaming.DATE_TIME,
            "individual": FileNaming.INDIVIDUAL,
            "counter": FileNaming.COUNTER,
        }
        if not string in stringEnumMap:
            raise ValueError("invalid string: " + string)
        return stringEnumMap.get(string)


class Pad4Mode(IntEnum):
    r"""
    Options for the PAD4 operating mode.

    All channels can be either voltage or current. Individual setting is not possible.
    """

    VOLTAGE = 0
    CURRENT = 1


class ThalesRemoteScriptWrapper(object):
    r"""
    Wrapper that uses the ThalesRemoteConnection class.
    The commands are explained in the `Remote2-Manual <https://doc.zahner.de/manuals/remote2.pdf>`_.
    In the document you can also find a table with error numbers which are returned.

    :param remoteConnection: The connection object to the Thales software.
    """

    undefindedStandardErrorString: str = ""
    _remote_connection: ThalesRemoteConnection

    def __init__(self, remoteConnection: ThalesRemoteConnection):
        self._remote_connection = remoteConnection
        try:
            versionReply = self.getThalesVersion(timeout=1)
        except TermConnectionError as err:
            raise ThalesRemoteError(
                "Please update the Thales software, it is too old for this package version."
            )

        if "devel" in versionReply:
            print("devel version")
        elif "" == versionReply:
            # timeout
            raise ThalesRemoteError(
                "Please update the Thales software, it is too old for this package version."
            )
        else:
            match = re.search(r"(\d+.\d+.\d+)", versionReply)
            versionString = match.group(1)
            thalesToOld = versiontuple(versionString) < versiontuple(
                MINIMUM_THALES_VERSION
            )
            if thalesToOld:
                raise ThalesRemoteError(
                    f"Please update the Thales software. This package requires at least version {MINIMUM_THALES_VERSION}, but version {versionString} is installed."
                )
        return

    def getCurrent(self) -> float:
        r"""
        Read the measured current from the device

        :returns: the measured current value
        """
        return self._requestValueAndParseUsingRegexp(
            "CURRENT", r"current=\s*(.*?)A?[\r\n]{0,2}$"
        )

    def getPotential(self) -> float:
        r"""
        Read the measured potential from the device

        :returns: the measured potential value
        """
        return self._requestValueAndParseUsingRegexp(
            "POTENTIAL", r"potential=\s*(.*?)V?[\r\n]{0,2}$"
        )

    def getVoltage(self) -> float:
        r"""
        Read the measured potential from the device

        :returns: the measured potential value
        """
        return self.getPotential()

    def setCurrent(self, current: float) -> str:
        r"""
        Set the output current

        :param current: the output current to set
        :returns: response string from the device
        """
        return self.setValue("Cset", current)

    def setPotential(self, potential: float) -> str:
        r"""
        Set the output potential

        :param potential: the output potential to set
        :returns: response string from the device
        """
        return self.setValue("Pset", potential)

    def setVoltage(self, potential) -> str:
        r"""
        Set the output potential

        :param potential: the output potential to set
        :returns: response string from the device
        """
        return self.setPotential(potential)

    def setMaximumShunt(self, shunt: int) -> str:
        r"""
        Set the maximum shunt index for measurement

        Set the maximum shunt index for impedance measurements.

        :param shunt: the number of the shunt
        :returns: response string from the device
        """
        return self.setValue("Rmax", shunt)

    def setMinimumShunt(self, shunt) -> str:
        r"""
        Set the minimum shunt for measurement

        Set the minimum shunt index for impedance measurements.

        :param shunt: index of the shunt to set
        :returns: response string from the device
        """
        return self.setValue("Rmin", shunt)

    def setShuntIndex(self, shunt: int) -> None:
        r"""
        Set the shunt index for measurement

        Fixes the shunt to the passed index.

        :param shunt: The number of the shunt.
        :returns: reponse string from the device
        """
        self.setMinimumShunt(shunt)
        self.setMaximumShunt(shunt)
        return

    def setVoltageRangeIndex(self, vrange: int) -> str:
        r"""
        Set the voltage range for measurement

        If a Zennium, Zennium E, Zennium E4 or a device from the IM6 series is
        used, the set index must match the U-buffer. If the U-buffer does not
        match the set value, the measurement is wrong.
        The Zennium pro, Zennium X and Zennium XC series automatically change
        the range.

        :param vrange: index of the voltage range
        :returns: response string from the device
        """
        return self.setValue("Potrange", vrange)

    def selectPotentiostat(self, device: int) -> str:
        r"""
        Select device onto which all subsequent calls to set* methods are forwarded

        First, the device must be selected. Only then can devices other than
        the internal main potentiostat be configured.

        :param device: Number of the device. 0 = Main. 1 = EPC channel 1 and so on.
        :returns: reponse string from the device
        """
        return self.setValue("DEV%", device)

    def selectPotentiostatWithoutPotentiostatStateChange(self, device: int) -> str:
        r"""
        Select device onto which all subsequent calls to set* methods are forwarded

        Device which is to be selected, on which the settings are output.
        First, the device must be selected.
        Only then can devices other than the internal main potentiostat be configured.
        The potentiostat is not turned off.

        :param device: Number of the device. 0 = Main. 1 = EPC channel 1 and so on.
        :returns: The response string from the device.
        """
        return self.setValue("DEVHOT%", device)

    def switchToSCPIControl(self) -> str:
        r"""
        Change away from operation as EPC device to SCPI operation

        This command works only with external potentiostats of the latest generation PP212, PP222, PP242 and XPOT2.
        After this command they are no longer accessible with the EPC interface.
        Then you can connect to the potentiostat with USB via the Comports.
        The change back to EPC operation is also done explicitly from the USB side.

        :returns: response string from the device
        """
        return self.executeRemoteCommand("SETUSB")

    def switchToSCPIControlWithoutPotentiostatStateChange(self) -> str:
        r"""
        Change away from operation as EPC device to SCPI operation.

        This command works only with external potentiostats of the latest generation XPOT2, PP2x2, EL1002.
        This requires a device firmware with at least version 1.0.4.
        After this command they are no longer accessible with the EPC interface.
        Then you can connect to the potentiostat with USB via the Comports.
        The change back to EPC operation is also done explicitly from the USB side.

        This function leaves the potentiostat in its current operating state and then switches to USB mode.
        This should only be used when it is really necessary to leave the potentiostat on,
        because between the change of control no quantities like current and voltage are monitored.

        To ensure that the switch between Thales and Python/SCPI is interference-free, the following procedure should be followed.
        This is necessary to ensure that both Thales and Python/SCPI have calibrated offsets, otherwise jumps may occur when switching modes:

         1. Connect Zennium with USB and EPC-device/power potentiostat (XPOT2, PP2x2, EL1002) with USB to the computer. As well as Zennium to power potentiostat by EPC cable.
         2. Switch on all devices.
         3. Allow the equipment to warm up for at least 30 minutes.
         4. Select and calibrate the EPC-device in Thales (with Remote2).
         5. Switching the EPC-device to SCPI mode via Remote2 command.
         6. Performing the offset calibration with Python/SCPI.
         7. Then it is possible to switch between Thales and Python/SCPI with the potentiostat switched on.

        With Thales, the DC operating point must first be set.
        When changing the EPC device then measures current and voltage and sets the size internally.
        When switching back to Thales, the same DC operating point must be set as when switching from Thales to USB.

        :returns: The response string from the device.
        """
        return self.executeRemoteCommand("HOT2USB")

    def getSerialNumber(self) -> str:
        r"""
        Get the serial number of the active device

        Active device ist the device selected with
        :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.selectPotentiostat`.

        :returns: the device serial number
        """
        reply = self.executeRemoteCommand("ALLNUM")
        match = re.search(r"(.*);(.*);([a-zA-Z]*)", reply)
        return match.group(2)

    def getDeviceInformation(self) -> tuple[str, str]:
        r"""
        Get the name and serial number of the active device

        :returns: tuple with the information about the selected potentiostat. (Name, Serialnumber).
        """
        reply = self.executeRemoteCommand("DEVINF")
        match = re.search(r"(.*);(.*);(.*);([0-9]*)", reply)
        return match.group(3), match.group(4)

    def getDeviceName(self) -> str:
        r"""
        Get the name of the active device

        :returns: the device name
        """
        reply = self.executeRemoteCommand("ALLNUM")
        match = re.search(r"(.*);(.*);([a-zA-Z]*)", reply)
        return match.group(3)

    def readSetup(self) -> str:
        r"""
        Read the currently set parameters

        A string containing the configuration is returned.
        For Example:

        .. code-block::

            OK;SETUP;Pset=1.0000e-05;Cset=1.0000e-06;Frq=1.0000e+03;Ampl=0.0000e+00;Nw=1;Pot=0;Gal=0;GAL=0;Cmin=-3.0000e+00;Cmax=3.0000e+00;Pmin=-5.2377e+00;Pmax=5.2377e+00;DEV=0;EPC=0;MAXDEV=4;ENDSETUP

        :returns: reponse string from the device
        """
        return self.executeRemoteCommand("SENDSETUP")

    def calibrateOffsets(self) -> str:
        r"""
        Perform offset calibration on the device

        When the instrument has warmed up for about 30 minutes,
        this command can be used to perform the offset calibration again.

        :returns: response string from the device
        """
        return self.executeRemoteCommand("CALOFFSETS")

    def enablePotentiostat(self, enabled: bool = True) -> str:
        r"""
        Switch the potentiostat on or off

        :param enabled: Switches the potentiostat on if True and off otherwise
        :returns: response string from the device
        """
        return self.executeRemoteCommand("Pot=" + ("-1" if enabled else "0"))

    def disablePotentiostat(self) -> str:
        r"""
        Switch the potentiostat off

        :returns: response string from the device
        """
        return self.enablePotentiostat(False)

    def setPotentiostatMode(self, potentiostatMode: PotentiostatMode) -> str:
        r"""
        Set the coupling of the potentiostat

        This can be PotentiostatMode.POTMODE_POTENTIOSTATIC or PotentiostatMode.POTMODE_GALVANOSTATIC or
        PotentiostatMode.POTMODE_PSEUDOGALVANOSTATIC.

        :param potentiostatMode: The coupling of the potentiostat
        :type potentiostatMode: :class:`~thales_remote.script_wrapper.PotentiostatMode`
        :returns: response string from the device
        """
        command_strings = {
            PotentiostatMode.POTMODE_POTENTIOSTATIC: "Gal=0:GAL=0",
            PotentiostatMode.POTMODE_GALVANOSTATIC: "Gal=-1:GAL=1",
            PotentiostatMode.POTMODE_PSEUDOGALVANOSTATIC: "Gal=0:GAL=-1",
        }
        if not potentiostatMode in command_strings:
            raise ValueError("invalid potentiostat mode: " + str(potentiostatMode))
        return self.executeRemoteCommand(command_strings.get(potentiostatMode))

    def enableRuleFileUsage(self, enable: bool = True) -> str:
        r"""
        Enable the usage of a rule file

        If the usage of the rule file is activated all the parameters required
        for the EIS, CV, and/or IE are taken from the rule file.
        The exact usage can be found in the remote manual.

        :param potentiostatMode: Enable the state of rule file usage.
        :returns: response string from the device
        """
        return self.setValue("UseRuleFile", 1 if enable else 0)

    def disableRuleFileUsage(self) -> str:
        r"""
        Disable the usage of a rule file

        :returns: reponse string from the device
        """
        return self.enableRuleFileUsage(False)

    """
    Section with PAD4 commands
    """

    def setupPad4Channel(
        self,
        card: int,
        channel: int,
        state: Union[int, bool],
        voltageRange: Optional[float] = None,
        shuntResistor: Optional[float] = None,
    ) -> None:
        r"""
        Setting a channel of a PAD4 card for an EIS measurement

        Each PAD4 channel must be configured individually. Each channel can be switched on or off individually.
        But the PAD4 measurements must still be switched on globally with :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.enablePad4Global`.
        Each channel can be given a different voltage range or shunt.

        :param card: index of the card starting at 1 and up to 4
        :param channel: index of the card starting at 1 and up to 4
        :param state: If `1` or `True` the channel is switched on else switched off
        :param voltageRange: input voltage range, if this differs from 4 V
        :param shuntResistor: shunt resistor, which is used. Only used if :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setupPad4ModeGlobal` is set to :class:`.Pad4Mode`.CURRENT
        :returns: reponse string from the device
        """
        commands = [f"PAD4={card};{channel};{1 if state else 0}"]
        if voltageRange:
            commands.append(f"PAD4_PRANGE={card};{channel};{voltageRange}")
        if shuntResistor:
            commands.append(f"PAD4_RSHUNT={card};{channel};{shuntResistor}")

        for command in commands:
            if isinstance(state, int):
                state = state == 1
            reply = self.executeRemoteCommand(command)

            if "ERROR" in reply:
                raise ThalesRemoteError(
                    reply.rstrip("\r")
                    + ThalesRemoteScriptWrapper.undefindedStandardErrorString
                )
        return

    def setupPad4ModeGlobal(self, mode: Pad4Mode) -> str:
        r"""
        Switch between current and voltage measurement

        The user can switch the type of PAD4 channels between voltage sense (standard configuration) and current sense (with additional shunt resistor).

        :param mode: :class:`.Pad4Mode`.VOLTAGE or :class:`.Pad4Mode`.CURRENT
        :returns: reponse string from the device
        """
        return self.executeRemoteCommand(f"PAD4MOD={mode.value}")

    def enablePad4Global(self, state: bool = True) -> str:
        r"""
        Switch on the set PAD4 channels

        The files of the measurement results with PAD4 are numbered consecutively.
        The lowest number is the main channel of the potentiostat.

        :param state: If state = True PAD4 channels are enabled.
        :returns: reponse string from the device
        """
        return self.setValue("PAD4ENA", 1 if state else 0)

    def disablePad4Global(self) -> str:
        r"""
        Switch off the set PAD4 channels

        :returns: reponse string from the device
        """
        return self.enablePad4Global(False)

    def readPad4SetupGlobal(self) -> str:
        r"""
        Read the currently set PAD4 parameters

        Reading the set PAD4 configuration.
        A string containing the configuration is returned.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("SENDPAD4SETUP")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    """
    Section with settings for EIS measurements.
    
    Additional informations can be found in the remote and EIS manual.
    """

    def setFrequency(self, frequency: float) -> str:
        r"""
        Set the output frequency for single frequency impedance.

        :param frequency: the output frequency for Impedance measurement to set
        :returns: reponse string from the device
        """
        return self.setValue("Frq", frequency)

    def setAmplitude(self, amplitude: float) -> str:
        r"""
        Set the output amplitude

        :param amplitude: the output amplitude for Impedance measurement to set in Volt or Ampere
        :returns: reponse string from the device
        """
        return self.setValue("Ampl", amplitude * 1e3)

    def setNumberOfPeriods(self, number_of_periods: Union[int, float]):
        r"""
        Set the number of periods to average for one impedance measurement

        :param number_of_periods: the number of periods / waves to average
        :returns: reponse string from the device
        """
        if isinstance(number_of_periods, float):
            number_of_periods = round(number_of_periods)
        if number_of_periods > 100:
            number_of_periods = 100

        if number_of_periods < 1:
            number_of_periods = 1

        return self.setValue("Nw", number_of_periods)

    def setUpperFrequencyLimit(self, frequency: float) -> str:
        r"""
        Set the upper frequency limit for EIS measurements

        :param frequency: the upper frequency limit to set
        :returns: reponse string from the device
        """
        return self.setValue("Fmax", frequency)

    def setLowerFrequencyLimit(self, frequency: float) -> str:
        r"""
        Set the lower frequency limit for EIS measurements

        :param frequency: the lower frequency limit to set
        :returns: reponse string from the device
        """
        return self.setValue("Fmin", frequency)

    def setStartFrequency(self, frequency: float) -> str:
        r"""
        Set the start frequency for EIS measurements

        :param frequency: the start frequency to set
        :returns: reponse string from the device
        """
        return self.setValue("Fstart", frequency)

    def setUpperStepsPerDecade(self, steps: int) -> str:
        r"""
        Set the number of steps per decade in frequency range above 66 Hz for EIS measurements

        :param steps: the number of steps per decade to set
        :returns: reponse string from the device
        """
        return self.setValue("dfm", steps)

    def setLowerStepsPerDecade(self, steps: int) -> str:
        r"""
        Set the number of steps per decade in frequency range below 66 Hz for EIS measurements

        :param steps: the number of steps per decade to set
        :returns: reponse string from the device
        """
        return self.setValue("dfl", steps)

    def setUpperNumberOfPeriods(self, periods: int) -> str:
        r"""
        Set the number of periods to measure in frequency range above 66 Hz for EIS measurements

        Note:
            value must be greater or equal than the one set with
            :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setLowerNumberOfPeriods`

        :param periods: the number of periods to set
        :returns: reponse string from the device
        """
        return self.setValue("Nws", periods)

    def setLowerNumberOfPeriods(self, periods: int) -> str:
        r"""
        Set the number of periods to measure in the frequency range at the lower frequency limit for EIS measurements.

        Note:
            value must be smaller or equal than the one set with
            :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setUpperNumberOfPeriods`.

        :param periods: the number of periods to set
        :returns: reponse string from the device
        """
        return self.setValue("Nwl", periods)

    def setScanStrategy(self, strategy: Union[ScanStrategy, str]) -> str:
        r"""
        Set the scan strategy for EIS measurements.

        strategy = "single": single sine.
        strategy = "multi": multi sine.
        strategy = "table": frequency table.

        :param strategy: the scan strategy to set for EIS measurements
        :returns: reponse string from the device
        """
        strat = strategy
        if isinstance(strategy, str):
            strat = ScanStrategy.stringToEnum(strategy)
        return self.setValue("ScanStrategy", strat.value)

    def setScanDirection(self, direction: Union[ScanDirection, str]) -> str:
        r"""
        Set the scan direction for EIS measurements.

        direction = "startToMax": Scan at first from start to maximum frequency.
        direction = "startToMin": Scan at first from start to lower frequency.

        :param direction: The scan direction for EIS measurements.
        :type direction: string
        :returns: reponse string from the device
        """
        dir = direction
        if isinstance(direction, str):
            dir = ScanDirection.stringToEnum(direction)
        return self.setValue("ScanDirection", dir.value)

    def getImpedance(
        self,
        frequency: Optional[float] = None,
        amplitude: Optional[float] = None,
        number_of_periods: Optional[int] = None,
    ) -> complex:
        r"""
        Measure the impedance

        Measure the impedance with the parameters. If the parameters are omitted the last will be used.

        :param frequency: The frequency to measure the impedance at.
        :param amplitude: The amplitude to measure the impedance with. In Volt if potentiostatic mode or Ampere for galvanostatic mode.
        :param number_of_periods: The number of periods / waves to average.
        :returns: Dict
        :rtype: complex
        """
        if frequency != None:
            self.setFrequency(frequency)

        if amplitude != None:
            self.setAmplitude(amplitude)

        if number_of_periods != None:
            self.setNumberOfPeriods(number_of_periods)

        reply = self.executeRemoteCommand("IMPEDANCE")

        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        match = re.search("impedance=\\s*(.*?),\\s*(.*?)$", reply)
        return complex(float(match.group(1)), float(match.group(2)))

    def getImpedanceAsArray(
        self,
        frequency: Optional[float] = None,
        amplitude: Optional[float] = None,
        number_of_periods: Optional[int] = None,
    ) -> List[float]:
        r"""
        Measure the impedance

        Measure the impedance with the parameters. If the parameters are omitted the last will be used.

        :param frequency: The frequency to measure the impedance at.
        :param amplitude: The amplitude to measure the impedance with. In Volt if potentiostatic mode or Ampere for galvanostatic mode.
        :param number_of_periods: The number of periods / waves to average.
        :returns: List wit [real, imag]
        :rtype: List
        """
        comp = self.getImpedance(frequency, amplitude, number_of_periods)
        return [comp.real, comp.imag]

    def getImpedancePad4(
        self,
        frequency: Optional[float] = None,
        amplitude: Optional[float] = None,
        number_of_periods: Optional[int] = None,
    ) -> dict[int, complex]:
        r"""
        Measure the impedance with PAD4 channels

        Measure the impedance and activated PAD4 channels with the parameters. If the parameters are omitted the last will be used.
        The method returns a dictionary containing the channels as keys and the complex impedance as value.
        The MAIN channel has index 0. Switched off PAD4 channels have an impedance of 0.

        :param frequency: The frequency to measure the impedance at.
        :param amplitude: The amplitude to measure the impedance with. In Volt if potentiostatic mode or Ampere for galvanostatic mode.
        :param number_of_periods: The number of periods / waves to average.
        :returns: A dictionary with channel index as key and the complex impedance as value.
        """
        if frequency != None:
            self.setFrequency(frequency)

        if amplitude != None:
            self.setAmplitude(amplitude)

        if number_of_periods != None:
            self.setNumberOfPeriods(number_of_periods)

        reply = self.executeRemoteCommand("PAD4IMP")

        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )

        matches = re.finditer(r"=\s*(?P<real>.*?),\s*(?P<imag>.*?)(?:;|$)", reply)
        return dict(
            (key, complex(float(val.group("real")), float(val.group("imag"))))
            for key, val in enumerate(matches)
        )

    def getImpedancePad4AsArray(
        self,
        frequency: Optional[float] = None,
        amplitude: Optional[float] = None,
        number_of_periods: Optional[int] = None,
    ) -> List[float]:
        r"""
        Measure the impedance with PAD4 channels

        Measure the impedance and activated PAD4 channels with the parameters. If the parameters are omitted the last will be used.
        The method returns a dictionary containing the channels as keys and the complex impedance as value.
        The MAIN channel has index 0. Switched off PAD4 channels have an impedance of 0.

        :param frequency: The frequency to measure the impedance at.
        :param amplitude: The amplitude to measure the impedance with. In Volt if potentiostatic mode or Ampere for galvanostatic mode.
        :param number_of_periods: The number of periods / waves to average.
        :returns: List wit [real, imag,...]
        :rtype: List
        """
        comp = self.getImpedancePad4(frequency, amplitude, number_of_periods)
        retval = []
        for [key, val] in comp.items():
            retval.append(val.real)
            retval.append(val.imag)
        return retval

    def setEISNaming(self, naming: Union[str, FileNaming]) -> str:
        r"""
        Set the EIS measurement naming rule.

        naming = "dateTime": extend the :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setEISOutputFileName` with date and time.
        naming = "counter": extend the :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setEISOutputFileName` with an sequential number.
        naming = "individual": the file is named like :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setEISOutputFileName`.

        :param naming: the EIS measurement naming rule to set.
        :returns: reponse string from the device
        """
        nameingType = naming
        if isinstance(naming, str):
            nameingType = FileNaming.stringToEnum(naming)
        return self.setValue("EIS_MOD", nameingType.value)

    def setEISCounter(self, number: int) -> str:
        r"""
        Set the current number of EIS measurement for filename.

        Current number for the file name for EIS measurements which is used next and then incremented.

        :param number: the next measurement number to set
        :returns: reponse string from the device
        """
        return self.setValue("EIS_NUM", number)

    def setEISOutputPath(self, path: str) -> str:
        r"""
        Set the path where the EIS measurements should be stored.

        The results must be stored on the C hard disk.
        If an error occurs test an alternative path or c:\THALES\temp.
        The directory must exist.

        :param path: path to the directory
        :returns: reponse string from the device
        """
        path = path.lower()  # c: has to be lower case
        self._checkForForbiddenCharactersInPath(path)
        return self.setValue("EIS_PATH", path)

    def setEISOutputFileName(self, name: str) -> str:
        r"""
        Set the basic EIS output filename.

        The basic name of the file, which is extended by a sequential number or the date and time.
        Only numbers, underscores and letters from a-Z may be used.

        If the name is set to "individual", the file with the same name must not yet exist.
        Existing files are not overwritten.

        :param name: basic name of the file
        :returns: reponse string from the device
        """
        self._checkForForbiddenCharactersInFilename(name)
        return self.setValue("EIS_ROOT", name)

    def measureEIS(self) -> str:
        r"""
        Make an EIS measurement.

        For the measurement all parameters must be specified before.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("EIS")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    """
    Section with settings for CV measurements.
    
    Additional informations can be found in the remote and CV manual.
    """

    def setCVStartPotential(self, potential: float) -> str:
        r"""
        Set the start potential of a CV measurement.

        :param potential: the start potential to set
        :returns: reponse string from the device
        """
        return self.setValue("CV_Pstart", potential)

    def setCVUpperReversingPotential(self, potential: float) -> str:
        r"""
        Set the upper reversal potential of a CV measurement.

        :param potential: the upper reversal potential to set
        :returns: reponse string from the device
        """
        return self.setValue("CV_Pupper", potential)

    def setCVLowerReversingPotential(self, potential: float) -> str:
        r"""
        Set the lower reversal potential of a CV measurement.

        :param potential: the lower reversal potential to set
        :returns: reponse string from the device
        """
        return self.setValue("CV_Plower", potential)

    def setCVEndPotential(self, potential: float) -> str:
        r"""
        Set the end potential of a CV measurement.

        :param potential: the end potential to set
        :returns: reponse string from the device
        """
        return self.setValue("CV_Pend", potential)

    def setCVStartHoldTime(self, time: float) -> str:
        r"""
        Setting the holding time at the start potential.

        The time must be given in seconds.

        :param time: the waiting time at start potential in s
        :returns: reponse string from the device
        """
        return self.setValue("CV_Tstart", time)

    def setCVEndHoldTime(self, time: float) -> str:
        r"""
        Setting the holding time at the end potential.

        The time must be given in seconds.

        :param time: the waiting time at end potential in s
        :returns: reponse string from the device
        """
        return self.setValue("CV_Tend", time)

    def setCVScanRate(self, scanRate: float) -> str:
        r"""
        Set the scan rate.

        The scan rate must be specified in V/s.

        :param scanRate: the scan rate to set in V/s
        :returns: reponse string from the device
        """
        return self.setValue("CV_Srate", scanRate)

    def setCVCycles(self, cycles: float) -> str:
        r"""
        Set the number of cycles.

        At least 0.5 cycles are necessary.
        The number of cycles must be a multiple of 0.5. 3.5 are also possible, for example.

        :param cycles: the number of CV cycles to set, at least 0.5.
        :returns: reponse string from the device
        """
        return self.setValue("CV_Periods", cycles)

    def setCVSamplesPerCycle(self, samples: int) -> str:
        r"""
        Set the number of measurements per CV cycle.

        :param samples: the number of measurments per cycle to set
        :returns: reponse string from the device
        """
        return self.setValue("CV_PpPer", samples)

    def setCVMaximumCurrent(self, current: float) -> str:
        r"""
        Set the maximum current.

        The maximum positive current at which the measurement is interrupted.
        This is also the current which is used to determine the shunt. This cannot be deactivated.

        :param current: The maximum current for measurement in A at which the measurement is interrupted
        :returns: reponse string from the device
        """
        return self.setValue("CV_Ima", current)

    def setCVMinimumCurrent(self, current: float) -> str:
        r"""
        Set the minimum current.

        The maximum negative current at which the measurement is interrupted.
        This is also the current which is used to determine the shunt. This cannot be deactivated.

        :param current: The minimum current for measurement in A.
        :returns: reponse string from the device
        """
        return self.setValue("CV_Imi", current)

    def setCVOhmicDrop(self, ohmicdrop: float) -> str:
        r"""
        Set the ohmic drop for CV measurement.

        :param ohmicdrop: The ohmic drop for measurement.
        :returns: reponse string from the device
        """
        return self.setValue("CV_Odrop", ohmicdrop)

    def enableCVAutoRestartAtCurrentOverflow(self, state: bool = True) -> str:
        r"""
        Automatically restart if current is exceeded.

        A new measurement is automatically started with a different
        reverse potential at which the current limit is not exceeded.

        :param state: If state == True the auto restart is enabled.
        :returns: reponse string from the device
        """
        return self.setValue("CV_AutoReStart", 1 if state else 0)

    def disableCVAutoRestartAtCurrentOverflow(self) -> str:
        r"""
        Disable automatic restart if current is exceeded.

        :returns: reponse string from the device
        """
        return self.enableCVAutoRestartAtCurrentOverflow(False)

    def enableCVAutoRestartAtCurrentUnderflow(self, state: bool = True) -> str:
        r"""
        Restart automatically if the current drops below the limit.

        A new measurement is automatically started with a smaller
        current range than that determined by the minimum and maximum current.

        :param state: If state = True the auto restart is enabled.
        :returns: reponse string from the device
        """
        return self.setValue("CV_AutoScale", 1 if state else 0)

    def disableCVAutoRestartAtCurrentUnderflow(self) -> str:
        r"""
        Disable automatically restart if the current drops below the limit.

        :returns: reponse string from the device
        """
        return self.enableCVAutoRestartAtCurrentUnderflow(False)

    def enableCVAnalogFunctionGenerator(self, state: bool = True) -> str:
        r"""
        Switch on the analog function generator (AFG).

        The analog function generator can only be used if it was purchased with the device.
        If the device has the AFG function, you will see a button in the CV software to activate this function.

        :param state: If state == True the AFG is used.
        :returns: reponse string from the device
        """
        return self.setValue("CV_AFGena", 1 if state else 0)

    def disableCVAnalogFunctionGenerator(self) -> str:
        r"""
        Disable the analog function generator (AFG).

        :returns: reponse string from the device
        """
        return self.enableCVAnalogFunctionGenerator(False)

    def setCVNaming(self, naming: Union[str, FileNaming]) -> str:
        r"""
        Set the CV measurement naming rule.

        naming = "dateTime": extend the :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setCVOutputFileName` with date and time.
        naming = "counter": extend the :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setCVOutputFileName` with an sequential number.
        naming = "individual": the file is named like :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setCVOutputFileName`.

        :param naming: CV measurement naming rule to set
        :returns: reponse string from the device
        """
        nameingType = naming
        if isinstance(naming, str):
            nameingType = FileNaming.stringToEnum(naming)
        return self.setValue("CV_MOD", nameingType.value)

    def setCVCounter(self, number: int) -> str:
        r"""
        Set the current number of CV measurement for filename.

        Current number for the file name for CV measurements which is used next and then incremented.

        :param number: the next measurement number
        :returns: reponse string from the device
        """
        return self.setValue("CV_NUM", number)

    def setCVOutputPath(self, path: str) -> str:
        r"""
        Set the path where the CV measurements should be stored.

        The results must be stored on the C hard disk.
        If an error occurs test an alternative path or c:\THALES\temp.
        The directory must exist.

        :param path: path to the output directory
        :returns: reponse string from the device
        """
        path = path.lower()  # c: has to be lower case
        self._checkForForbiddenCharactersInPath(path)
        return self.setValue("CV_PATH", path)

    def setCVOutputFileName(self, name: str) -> str:
        r"""
        Set the basic CV output filename.

        The basic name of the file, which is extended by a sequential number or the date and time.
        Only numbers, underscores and letters from a-Z may be used.

        If the name is set to "individual", the file with the same name must not yet exist.
        Existing files are not overwritten.

        :param name: basic name of the file to set
        :returns: reponse string from the device
        """
        self._checkForForbiddenCharactersInFilename(name)
        return self.setValue("CV_ROOT", name)

    def checkCVSetup(self) -> str:
        r"""
        Check the set parameters.

        With the error number the wrong parameter can be found.
        The error numbers are listed in the Remote2 manual.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("CHECKCV")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def readCVSetup(self) -> str:
        r"""
        Read the set parameters.

        After checking with checkCVSetup() the parameters can be read back from the workstation.
        This command returns a list of all set parameters.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("SENDCVSETUP")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def measureCV(self) -> str:
        r"""
        Make a CV (cyclic voltammetry) measurement.

        Note:
            Before starting the measurement, all parameters must be checked with
            :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.checkCVSetup`.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("CV")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    """
    Section with settings for IE measurements.
    
    Additional informations can be found in the remote and IE manual.
    """

    def setIEFirstEdgePotential(self, potential: float) -> str:
        r"""
        Set the first edge potential.

        :param potential: The potential of the first edge in V.
        :returns: reponse string from the device
        """
        return self.setValue("IE_EckPot1", potential)

    def setIESecondEdgePotential(self, potential: float) -> str:
        r"""
        Set the second edge potential.

        :param potential: The potential of the second edge in V.
        :returns: reponse string from the device
        """
        return self.setValue("IE_EckPot2", potential)

    def setIEThirdEdgePotential(self, potential: float) -> str:
        r"""
        Set the third edge potential.

        :param potential: The potential of the third edge in V.
        :returns: reponse string from the device
        """
        return self.setValue("IE_EckPot3", potential)

    def setIEFourthEdgePotential(self, potential: float) -> str:
        r"""
        Set the fourth edge potential.

        :param potential: The potential of the fourth edge in V.
        :returns: reponse string from the device
        """
        return self.setValue("IE_EckPot4", potential)

    def setIEFirstEdgePotentialRelation(self, relation: float) -> str:
        r"""
        Set the relation of the first edge potential.

        relation = "absolute": Absolute relation of the potential.
        relation = "relative": Relative relation of the potential.

        :param relation: The relation of the edge potential absolute or relative.
        :returns: reponse string from the device
        """
        if relation == "relative":
            relation = -1
        else:
            relation = 0
        return self.setValue("IE_EckPot1rel", relation)

    def setIESecondEdgePotentialRelation(self, relation) -> str:
        r"""
        Set the relation of the second edge potential.

        relation = "absolute": Absolute relation of the potential.
        relation = "relative": Relative relation of the potential.

        :param relation: The relation of the edge potential absolute or relative.
        :returns: reponse string from the device
        """
        if relation == "relative":
            relation = -1
        else:
            relation = 0
        return self.setValue("IE_EckPot2rel", relation)

    def setIEThirdEdgePotentialRelation(self, relation) -> str:
        r"""
        Set the relation of the third edge potential.

        relation = "absolute": Absolute relation of the potential.
        relation = "relative": Relative relation of the potential.

        :param relation: The relation of the edge potential absolute or relative.
        :returns: reponse string from the device
        """
        if relation == "relative":
            relation = -1
        else:
            relation = 0
        return self.setValue("IE_EckPot3rel", relation)

    def setIEFourthEdgePotentialRelation(self, relation) -> str:
        r"""
        Set the relation of the fourth edge potential.

        relation = "absolute": Absolute relation of the potential.
        relation = "relative": Relative relation of the potential.

        :param relation: The relation of the edge potential absolute or relative.
        :returns: reponse string from the device
        """
        if relation == "relative":
            relation = -1
        else:
            relation = 0
        return self.setValue("IE_EckPot4rel", relation)

    def setIEPotentialResolution(self, resolution: float) -> str:
        r"""
        Set the potential resolution.

        The potential step size for IE measurements in V.

        :param relation: The resolution for the measurement in V.
        :returns: reponse string from the device
        """
        return self.setValue("IE_Resolution", resolution)

    def setIEMinimumWaitingTime(self, time: float) -> str:
        r"""
        Set the minimum waiting time.

        The minimum waiting time on each step of the IE measurement.
        This time is at least waited, even if the tolerance abort criteria are met.

        :param time: The waiting time in seconds.
        :returns: reponse string from the device
        """
        return self.setValue("IE_WZmin", time)

    def setIEMaximumWaitingTime(self, time: float) -> str:
        r"""
        Set the maximum waiting time.

        The maximum waiting time on each step of the IE measurement.
        After this time the measurement is stopped at this potential
        and continued with the next potential even if the tolerances are not reached.

        :param time: The waiting time in seconds.
        :returns: reponse string from the device
        """
        return self.setValue("IE_WZmax", time)

    def setIERelativeTolerance(self, tolerance: float) -> str:
        r"""
        Set the relative tolerance criteria.

        This parameter is only used in sweep mode steady state.
        The relative tolerance to wait in percent.
        The explanation can be found in the IE manual.

        :param tolerance: The tolerance to wait until break, 0.01 = 1%.
        :returns: reponse string from the device
        """
        return self.setValue("IE_Torel", tolerance)

    def setIEAbsoluteTolerance(self, tolerance: float) -> str:
        r"""
        Set the absolute tolerance criteria.

        This parameter is only used in sweep mode steady state.
        The absolute tolerance to wait in A.
        The explanation can be found in the IE manual.

        :param tolerance: The tolerance to wait until break, 0.01 = 1%.
        :returns: reponse string from the device
        """
        return self.setValue("IE_Toabs", tolerance)

    def setIEOhmicDrop(self, ohmicdrop: float) -> str:
        r"""
        Set the ohmic drop for IE measurement.

        :param ohmicdrop: The ohmic drop for measurement.
        :returns: reponse string from the device
        """
        return self.setValue("IE_Odrop", ohmicdrop)

    def setIESweepMode(self, mode) -> str:
        r"""
        Set the sweep mode.

        The explanation of the modes can be found in the IE manual.
        mode="steady state"
        mode="fixed sampling"
        mode="dynamic scan"

        :param mode: The sweep mode for measurement.
        :returns: reponse string from the device
        """
        if mode == "steady state":
            mode = 0
        elif mode == "dynamic scan":
            mode = 2
        else:
            mode = 1
        return self.setValue("IE_SweepMode", mode)

    def setIEScanRate(self, scanRate: float) -> str:
        r"""
        Set the scan rate.

        This parameter is only used in sweep mode dynamic scan.
        The scan rate must be specified in V/s.

        :param scanRate: The scan rate in V/s.
        :returns: reponse string from the device
        """
        return self.setValue("IE_Srate", scanRate)

    def setIEMaximumCurrent(self, current: float) -> str:
        r"""
        Set the maximum current.

        The maximum positive current at which the measurement is interrupted.

        :param current: The maximum current for measurement in A.
        :returns: reponse string from the device
        """
        return self.setValue("IE_Ima", current)

    def setIEMinimumCurrent(self, current: float) -> str:
        r"""
        Set the minimum current.

        The maximum negative current at which the measurement is interrupted.

        :param current: The minimum current for measurement in A.
        :returns: reponse string from the device
        """
        return self.setValue("IE_Imi", current)

    def setIENaming(self, naming: Union[str, FileNaming]) -> str:
        r"""
        Set the IE measurement naming rule.

        naming = "dateTime": extend the :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setIEOutputFileName` with date and time.
        naming = "counter": extend the :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setIEOutputFileName` with an sequential number.
        naming = "individual": the file is named like :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.setIEOutputFileName`.

        :param naming: The IE measurement naming rule.
        :type naming: string
        :returns: reponse string from the device
        """
        nameingType = naming
        if isinstance(naming, str):
            nameingType = FileNaming.stringToEnum(naming)
        return self.setValue("IE_MOD", nameingType.value)

    def setIECounter(self, number: int) -> str:
        r"""
        Set the current number of IE measurement for filename.

        Current number for the file name for EIS measurements which is used next and then incremented.

        :param number: The next measurement number.
        :returns: reponse string from the device
        """
        return self.setValue("IE_NUM", number)

    def setIEOutputPath(self, path: str) -> str:
        r"""
        Set the path where the IE measurements should be stored.

        The results must be stored on the C hard disk. If an error occurs test an alternative path or c:\THALES\temp.
        The directory must exist.

        :param path: The path to the directory.
        :returns: reponse string from the device
        """
        path = path.lower()  # c: has to be lower case
        self._checkForForbiddenCharactersInPath(path)
        return self.setValue("IE_PATH", path)

    def setIEOutputFileName(self, name: str) -> str:
        r"""
        Set the basic IE output filename.

        The basic name of the file, which is extended by a sequential number or the date and time.
        Only numbers, underscores and letters from a-Z may be used.

        If the name is set to "individual", the file with the same name must not yet exist.
        Existing files are not overwritten.

        :param name: The basic name of the file.
        :returns: reponse string from the device
        """
        self._checkForForbiddenCharactersInFilename(name)
        return self.setValue("IE_ROOT", name)

    def checkIESetup(self) -> str:
        r"""
        Check the set parameters.

        With the error number the wrong parameter can be found.
        The error numbers are listed in the Remote2 manual.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("CHECKIE")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def readIESetup(self) -> str:
        r"""
        Read the set parameters.

        After checking with :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.checkIESetup` the parameters can be read back from the workstation.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("SENDIESETUP")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def measureIE(self) -> str:
        r"""
        Measure IE.

        Before measurement, all parameters must be checked with :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.checkIESetup`.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("IE")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    """
    Section of remote functions for the sequencer.
 
    With the sequencer DC profiles can be described textually.
    For instructions on how the sequencer file is structured, please refer to the manual of the sequencer.
    """

    def selectSequence(self, number: int) -> str:
        r"""
        Select the sequence to run with runSequence().

        **If this method gives an error message that the sequence does not exist, then check the ZTrace.
        This error message can also mean that a wrong character occurs or a parameter is out of range.**

        The sequences must be stored under "C:\THALES\script\sequencer\sequences".
        Sequences from 0 to 9 can be created.
        These must have the names from "sequence00.seq" to "sequence09.seq".

        :param number: The number of the sequence.
        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("SELSEQ=" + str(number))
        if reply != "SELOK\r":
            raise ThalesRemoteError(reply.rstrip("\r") + " Check ZTrace.")
        return reply

    def setSequenceNaming(self, naming: Union[str, FileNaming]) -> str:
        r"""
        Set the sequence measurement naming rule.

        naming = "dateTime": extend the setSequenceOutputFileName(name) with date and time.
        naming = "counter": extend the setSequenceOutputFileName(name) with an sequential number.
        naming = "individual": the file is named like setEISOutputFileName(name).

        :param naming: The naming rule.
        :type naming: string
        :returns: reponse string from the device
        """
        nameingType = naming
        if isinstance(naming, str):
            nameingType = FileNaming.stringToEnum(naming)
        return self.setValue("SEQ_MOD", nameingType.value)

    def setSequenceCounter(self, number: int) -> str:
        r"""
        Set the current number of sequence measurement for filename.

        Current number for the file name for EIS measurements which is used next and then incremented.

        :param number: The next measurement number
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_NUM", number)

    def setSequenceOutputPath(self, path: str) -> str:
        r"""
        Set the path where the sequence measurements should be stored.

        The results must be stored on the C hard disk. If an error occurs test an alternative path or c:\THALES\temp.
        The directory must exist.

        :param path: The path to the directory.
        :returns: reponse string from the device
        """
        path = path.lower()  # c: has to be lower case
        self._checkForForbiddenCharactersInPath(path)
        return self.setValue("SEQ_PATH", path)

    def setSequenceOutputFileName(self, name: str) -> str:
        r"""
        Set the basic sequence output filename.

        The basic name of the file, which is extended by a sequential number or the date and time.
        Only numbers, underscores and letters from a-Z may be used.

        If the name is set to "individual", the file with the same name must not yet exist.
        Existing files are not overwritten.

        :param name: The basic name of the file.
        :returns: reponse string from the device
        """
        self._checkForForbiddenCharactersInFilename(name)
        return self.setValue("SEQ_ROOT", name)

    def enableSequenceAcqGlobal(self, state: bool = True) -> str:
        r"""
        Switch on that up to 8 ACQ channels are used in the sequencer.

        In the ACQ setup, the display channels must be set before they can be allowed.
        Then the individual channels must be activated with a separate command.

        :param state: If state = True ACQ channels are enabled.
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_ACQENA", -1 if state else 0)

    def disableSequenceAcqGlobal(self) -> str:
        r"""
        Switch off that up to 8 ACQ channels are used in the sequencer.

        :returns: reponse string from the device
        """
        return self.enableSequenceAcqGlobal(False)

    def enableSequenceAcqChannel(self, channel: int, state: bool = True) -> str:
        r"""
        Switch on the set sequence ACQ channel

        The channel index for the ACQ channel results from the ACQ setup index plus 1.
        This means that in the ACQ, for example, the set display channel 2 becomes channel 3 in the sequencer.

        :param channel: index of the ACQ channel.
        :param state: If state = True ACQ channel is enabled.
        :returns: reponse string from the device
        """
        command = f"SEQ_ACQ={channel};{1 if state else 0}"
        reply = self.executeRemoteCommand(command)

        if "ERROR" in reply:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def disableSequenceAcqChannel(self, channel) -> str:
        r"""
        Switch off the set sequence ACQ channel

        :param channel: index of the ACQ channel.
        :returns: reponse string from the device
        """
        return self.enableSequenceAcqChannel(channel, False)

    def readSequenceAcqSetup(self) -> str:
        r"""
        Read the currently set sequencer ACQ Setup

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("SENDSEQACQSETUP")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def runSequence(self) -> str:
        r"""
        Run the selected sequence.

        This command executes the selected sequence between 0 and 9.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("DOSEQ")
        if reply != "SEQ DONE\r":
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def runSequenceFile(
        self,
        filepath: str,
        sequence_folder: str = "C:/THALES/script/sequencer/sequences",
        sequence_number: int = 9,
    ) -> str:
        r"""
        Run the sequence at filepath.

        The file from the specified path is copied as sequence sequence_number=9 to the correct location in the Thales directory and then selected and executed.
        The default sequence number is 9 and does not need to be changed.
        The old sequence sequence_number is overwritten.

        The variable sequence_folder is ONLY NECESSARY if python is running on ANOTHER COMPUTER like the one connected to the Zennium.
        For this the controlling computer must have access to the hard disk of the computer with the Zennium to access the THALES folder.
        In this case the path to the sequences folder in sequence_folder must be set to "C:/THALES/script/sequencer/sequences" on the computer with the zennium.

        :param filepath: Filepath of the sequence.
        :param sequence_folder: The filepath to the THALES sequence folder.
            Does not normally need to be transferred and changed. Explanation see in the text before.
        :param sequence_number: The number in the THALES sequence directory.
            Does not normally need to be transferred and changed.
        :returns: reponse string from the device
        """
        if sequence_number > 9 or sequence_number < 0:
            raise ThalesRemoteError("Wrong sequence number.")

        sequence_folder = sequence_folder + "/sequence{:02d}.seq".format(
            sequence_number
        )

        if filepath.find(".seq") < 0:
            raise ThalesRemoteError("Wrong sequence file extension.")
        if os.path.exists(sequence_folder):
            os.remove(sequence_folder)
        shutil.copy2(filepath, sequence_folder)

        self.selectSequence(sequence_number)
        return self.runSequence()

    def setSequenceOhmicDrop(self, value: float) -> str:
        r"""
        Set the Ohmic Drop or IR Drop for the sequencer.

        0 to 1 tera ohm.

        Parameter name in sequence file: ``rodrop``

        :param value: Ohmic Drop in Ohm
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_RODROP", value)

    def setSequenceMaximumRuntime(self, value: float) -> str:
        r"""
        Set the maximum runtime in hours

        0.1 to 1000 hours.

        Parameter name in sequence file: ``max_ti``

        :param value: maximum runtime in hours
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_MAXTIME", value)

    def setSequenceUpperPotentialLimit(self, value: float) -> str:
        r"""
        Set the upper potential limit

        Parameter name in sequence file: ``pot_hi``

        :param value: upper potential limit
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_EUPPER", value)

    def setSequenceLowerPotentialLimit(self, value: float) -> str:
        r"""
        Set the lower potential limit

        Parameter name in sequence file: ``pot_lo``

        :param value: lower potential limit
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_ELOWER", value)

    def setSequenceUpperCurrentLimit(self, value: float) -> str:
        r"""
        Set the upper current limit

        Parameter name in sequence file: ``cur_hi``

        :param value: upper current limit
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_IUPPER", value)

    def setSequenceLowerCurrentLimit(self, value: float) -> str:
        r"""
        Set the lower current limit

        Parameter name in sequence file: ``cur_lo``

        :param value: lower current limit
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_ILOWER", value)

    def setSequenceCurrentRange(self, value: float) -> str:
        r"""
        Set the current range

        Parameter name in sequence file: ``cur_ra``

        :param value: current range in ampere.
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_IRANGE", value)

    def setSequencePotentialLatencyWindow(self, value: float) -> str:
        r"""
        Set the potential latency window

        This defines what happens when a limit value is exceeded.
        The values from the following table are possible.

        .. list-table:: Latency Window Values
            :widths: 10 40
            :header-rows: 1

            * - Value
              - Meaning
            * - -1
              - potentiostat is not switched off
            * - t
              - potentiostat is switched off after t seconds, maximum 30
            * - 2
              - potentiostat is switched off after 2 seconds
            * - 0
              - potentiostat is switched off immediately

        Parameter name in sequence file: ``pot_of``

        :param value: values meaning see in the table
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_POTOFLO", value)

    def setSequenceCurrentLatencyWindow(self, value: float) -> str:
        r"""
        Set the current latency window

        This defines what happens when a limit value is exceeded.
        The values from the following table are possible.

        .. list-table:: Latency Window Values
            :widths: 10 40
            :header-rows: 1

            * - Value
              - Meaning
            * - -1
              - potentiostat is not switched off
            * - t
              - potentiostat is switched off after t seconds, maximum 30
            * - 2
              - potentiostat is switched off after 2 seconds
            * - 0
              - potentiostat is switched off immediately

        Parameter name in sequence file: ``cur_of``

        :param value: values meaning see in the table
        :returns: reponse string from the device
        """
        return self.setValue("SEQ_CUROFLO", value)

    """
    Section with commands for FRA mode
    """

    def enableFraMode(self, state: bool = True) -> str:
        r"""
        Enables the use of the FRA/FRA-X probe.

        With the FRA Probe, external power potentiostats, signal generators, sources, sinks can be
        controlled analog for impedance measurements.

        `FRA Product Page <https://zahner.de/products-details/probes/fra-probe>`_

        `FRA Manual <https://doc.zahner.de/hardware/fra_probe.pdf>`_

        Before this function is called, the analog interface to the external interface must be initialized
        with the correct factors. It may be necessary to use + or - as sign, this must be tested to ensure
        that potentiostatic and galvanostatic function correctly. The set factors should also be verified
        by measuring with a multimeter whether the gains and offsets have been set correctly, as these are
        not trivial to determine.

        :param state: If state = True FRA mode is enabled.
        :returns: reponse string from the device
        """
        if state == True:
            state = 1
        else:
            state = 0
        return self.setValue("FRA", state)

    def readFraSetup(self) -> str:
        r"""
        Read the currently set FRA parameters/configuration.

        A string containing the configuration is returned.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("SENDFRASETUP")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def disableFraMode(self) -> str:
        r"""
        Disables the use of the FRA/FRA-X probe.

        After disabling FRA mode, all parameters must be re-set for use with EPC devices, for example.

        :returns: reponse string from the device
        """
        return self.enableFraMode(False)

    def setFraVoltageInputGain(self, value: float) -> str:
        r"""
        Sets the input voltage gain of the FRA device.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_POT_IN", value)

    def setFraVoltageInputOffset(self, value: float) -> str:
        r"""
        Sets the input voltage offset of the FRA device.

        This value must be determined empirically as it depends on the gains used.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_POT_IN_OFF", value)

    def setFraVoltageOutputGain(self, value: float) -> str:
        r"""
        Sets the output voltage gain of the FRA device.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_POT_OUT", value)

    def setFraVoltageOutputOffset(self, value: float) -> str:
        r"""
        Sets the output voltage offset of the FRA device.

        This value must be determined empirically as it depends on the gains used.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_POT_OUT_OFF", value)

    def setFraVoltageMinimum(self, value: float) -> str:
        r"""
        Sets the minimum voltage of the FRA device.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_POT_MIN", value)

    def setFraVoltageMaximum(self, value: float) -> str:
        r"""
        Sets the maximum voltage of the FRA device.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_POT_MAX", value)

    def setFraCurrentInputGain(self, value: float) -> str:
        r"""
        Sets the input current gain of the FRA device.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_CUR_IN", value)

    def setFraCurrentInputOffset(self, value: float) -> str:
        r"""
        Sets the input current offset of the FRA device.

        This value must be determined empirically as it depends on the gains used.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_CUR_IN_OFF", value)

    def setFraCurrentOutputGain(self, value: float) -> str:
        r"""
        Sets the output current gain of the FRA device.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_CUR_OUT", value)

    def setFraCurrentOutputOffset(self, value: float) -> str:
        r"""
        Sets the output current offset of the FRA device.

        This value must be determined empirically as it depends on the gains used.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_CUR_OUT_OFF", value)

    def setFraCurrentMinimum(self, value: float) -> str:
        r"""Sets the minimum current of the FRA device.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_CUR_MIN", value)

    def setFraCurrentMaximum(self, value: float) -> str:
        r"""Sets the maximum current of the FRA device.

        :param value: The value to set.
        :returns: reponse string from the device
        """
        return self.setValue("FRA_CUR_MAX", value)

    def setFraPotentiostatMode(self, potentiostatMode: PotentiostatMode) -> str:
        r"""
        Set the coupling of the FRA mode.

        This can be PotentiostatMode.POTMODE_POTENTIOSTATIC or PotentiostatMode.POTMODE_GALVANOSTATIC

        :param potentiostatMode: The coupling of the FRA mode
        :returns: reponse string from the device
        """
        if potentiostatMode == PotentiostatMode.POTMODE_POTENTIOSTATIC:
            command = "FRAGAL=0"
        elif potentiostatMode == PotentiostatMode.POTMODE_GALVANOSTATIC:
            command = "FRAGAL=1"
        else:
            return ValueError(
                "PotentiostatMode.POTMODE_POTENTIOSTATIC or PotentiostatMode.POTMODE_GALVANOSTATIC"
            )
        return self.executeRemoteCommand(command)

    """
    Section with methods for the ACQ channels
    """

    def readAcqSetup(self) -> str:
        r"""
        Read the currently set ACQ parameters

        Reading the set FRA configuration.
        A string containing the configuration is returned.

        :returns: reponse string from the device
        """
        reply = self.executeRemoteCommand("SENDACQSETUP")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip("\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def readAllAcqChannels(self) -> dict[str, float]:
        r"""
        Read all active ACQ channels.

        A string is read from Thales which contains the ACQ channels as follows.
        .. code-block::

            ACQVAL(0)= 2.632052e-01;ACQVAL(1)= 8.413594e-02;ACQVAL(2)= 5.338292e+01

        :returns: Dict which contains the ACQ channel index as key and the value as value.
        """
        reply = self.executeRemoteCommand("ANALOGALL")
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip(r"\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )

        result_dict = {}

        pairs = reply.split(";")

        for pair in pairs:
            key = re.search(r"\((.*?)\)", pair).group(1)
            value = re.search(r"=[\s]*(.*)", pair).group(1)
            result_dict[int(key)] = float(value)

        return result_dict

    def readAcqChannel(self, channel: int) -> float:
        r"""
        Read the ACQ channel to the passed index.

        :param channel: index of the ACQ channel
        :returns: value of the channel
        """
        self.setValue("CHANNEL", channel)
        value = self._requestValueAndParseUsingRegexp("ANALOGIN", r"=[\s]*(.*)")
        return value

    def disableAcq(self) -> str:
        r"""
        Disable all configured ACQ channels.

        :returns: reponse string from the device
        """
        return self.enableAcq(False)

    def enableAcq(self, enabled: bool = True) -> str:
        r"""
        Enable configured ACQ channels.

        With this command, the ACQ channels can only be used for the EIS, CV and IE methods.
        The ACQ channels must first be set up via the GUI.

        :param enabled: True to enable ACQ.
        :returns: response string from the device
        """
        return self.executeRemoteCommand("ACQENA=" + ("1" if enabled else "0"))

    """
    Internal basic functions, which are used by the other methods.
    Could be private. But if a remote command is not implemented, you can implement it yourself.
    """

    def setValue(self, name: str, value: Union[int, float, str, Any]) -> str:
        r"""
        Set an Remote2 parameter or value.

        :param name: name of the Remote2 parameter
        :param value: value of the parameter to set
        :returns: response string from the device
        """
        if isinstance(value, float):
            value = "{:.14e}".format(value)

        reply = self.executeRemoteCommand(name + "=" + str(value))
        if "ERROR" in reply:
            raise ThalesRemoteError(
                reply.rstrip(r"\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        return reply

    def executeRemoteCommand(self, command: str) -> str:
        r"""
        Directly execute a query to Remote Script.

        :param command: The command query string, e.g. "IMPEDANCE" or "Pset=0".
        :returns: reponse string from the device
        """
        return self._remote_connection.sendStringAndWaitForReplyString(
            "1:" + command + ":", 2
        )

    def forceThalesIntoRemoteScript(self) -> str:
        r"""
        Prompts Thales to start the Remote Script.

        Will switch a running Thales from anywhere like the main menu after
        startup to running measurements into Remote Script so it can process
        further requests.

        .. note::
            This happens rather quickly if Thales is idle in main menu but
            may take some time if it's performing an EIS measurement or doing
            something else. For high stability applications 20 seconds would
            probably be a save bet.

        :returns: reponse string from the device
        """
        self._remote_connection.sendStringAndWaitForReplyString(
            f"3,{self._remote_connection.getConnectionName()},0,OFF", 128
        )
        return self._remote_connection.sendStringAndWaitForReplyString(
            f"2,{self._remote_connection.getConnectionName()}", 128
        )

    def hideWindow(self):
        r"""
        Hide Thales window.

        This is for remote integrations to prevent operation of the GUI.
        Thales is still visible in the taskbar but can no longer be maximized.

        :returns: reponse string from the device
        """
        return self._remote_connection.sendStringAndWaitForReplyString(
            f"3,{self._remote_connection.getConnectionName()},5,off", 128
        )

    def showWindow(self):
        r"""
        Show Thales window.

        If the Thales window has been hidden, it can be displayed again with this command.

        :returns: reponse string from the device
        """
        return self._remote_connection.sendStringAndWaitForReplyString(
            f"3,{self._remote_connection.getConnectionName()},5,on", 128
        )

    def getThalesVersion(self, timeout: Optional[float] = None):
        r"""
        Get Thales Version.

        Reads the current Thales version. Same version as in the application title line.

        :param timeout: The time in seconds in which the term must provide the answer.
        :returns: serialnumber as string.
        """
        return self._remote_connection.sendStringAndWaitForReplyString(
            f"3,{self._remote_connection.getConnectionName()},7", 128, timeout
        ).split(",")[2]

    def getWorkstationHeartBeat(self, timeout: Optional[float] = None) -> float:
        r"""
        Query the heartbeat time from the Term software for the workstation and the Thales software accordingly.

        The complete documentation can be found in the `DevCli-Manual <https://doc.zahner.de/manuals/devcli.pdf>`_ Page 8.


        The timeout is not set by default and the command blocks indefinitely.
        However, a time in seconds can optionally be specified for the timeout.
        When the timeout expires, an exception is thrown by the socket.

        :param timeout: The time in seconds in which the term must provide the answer.
        :returns: The HeartBeat time in milli seconds.
        """
        retval = self._remote_connection.sendStringAndWaitForReplyString(
            f"1,{self._remote_connection.getConnectionName()}", 128, timeout
        )
        return float(retval.split(",")[2])

    def getSerialNumberFromTerm(self) -> str:
        r"""
        Get the serial number of the workstation via the Term software.

        The serial number of the active potentiostat with EPC devices can be read with the
        :func:`~thales_remote.script_wrapper.ThalesRemoteScriptWrapper.getSerialNumber` function.
        This function returns only the serial number of the workstation, which is determined by the Term software.

        :returns: The workstation serial number.
        """
        retval = self._remote_connection.sendStringAndWaitForReplyString(
            f"3,{self._remote_connection.getConnectionName()},6", 128
        )
        return retval.split(",")[2]

    def getTermIsActive(self, timeout: float = 2) -> bool:
        r"""
        Check if the Term still responds to requests.

        Whether the term is still active is checked by sending a heartbeat command.
        If the term does not respond after the timeout, an exception is thrown and this is caught with the except block.
        The time returned by getWorkstationHeartBeat is not relevant,
        it is only important that a response was returned within the time.

        Afterwards False is returned if the exception was resolved.
        Once False is returned, the connection to the term MUST be completely re-established with
        connectToTerm and forceThalesIntoRemoteScript.
        The workstation must also be reset. To re-establish a controlled state.

        :param timeout: Time in seconds in which the term must provide the answer, default 2 seconds.
        :returns: True or False if the Term is Active.
        """
        active = True
        try:
            self._remote_connection.sendStringAndWaitForReplyString(
                f"1,{self._remote_connection.getConnectionName()}", 128, timeout
            )
        except:
            active = False
        return active

    """
    The following methods should not be called by the user.
    They are marked with the prefix '_' after the Python convention for proteced.
    """

    def _requestValueAndParseUsingRegexp(self, command: str, pattern: str) -> float:
        reply = self.executeRemoteCommand(command)
        if reply.find("ERROR") >= 0:
            raise ThalesRemoteError(
                reply.rstrip(r"\r")
                + ThalesRemoteScriptWrapper.undefindedStandardErrorString
            )
        match = re.search(pattern, reply)
        return float(match.group(1))

    def _checkForForbiddenCharactersInPath(self, string: str):
        r"""
        This function ensures that only the permitted characters appear in the path to:
        https://doc.zahner.de/manuals/remote2.pdf

        :param string: string to check.
        """
        pattern = re.compile(r"(^[a-k]{1}[:]{1}[\/\\0-9a-zA-Z_\+-]+$)")
        match = pattern.fullmatch(string)
        if match is None:
            raise ValueError(f'invalid path: "{string}"')
        return

    def _checkForForbiddenCharactersInFilename(self, string: str):
        r"""
        This function ensures that only the permitted characters appear in the filename to:
        https://doc.zahner.de/manuals/remote2.pdf

        :param string: string to check.
        """
        pattern = re.compile(r"(^[0-9a-zA-Z_\+-]+$)")
        match = pattern.fullmatch(string)
        if match is None:
            raise ValueError(f'invalid filename: "{string}"')
        return



import sys
from thales_remote.connection import ThalesRemoteConnection
from thales_remote.script_wrapper import PotentiostatMode, ThalesRemoteScriptWrapper
import time
import threading

zenniumConnectionLiveData = None

keepThreadRunning = True


def watchThreadFunction():
    global keepThreadRunning

    zenniumConnection = ThalesRemoteConnection()
    zenniumConnection.connectToTerm("localhost", "Watch")

    zahnerZennium = ThalesRemoteScriptWrapper(zenniumConnection)

    while keepThreadRunning:
        time.sleep(1)
        active = zahnerZennium.getTermIsActive()
        print("active state: " + str(active))
        if active:
            print("beat count: " + str(zahnerZennium.getWorkstationHeartBeat()))
    zenniumConnection.disconnectFromTerm()
    return


def liveDataThreadFunction():
    global keepThreadRunning
    global zenniumConnectionLiveData

    print("live thread started")
    while keepThreadRunning:
        try:
            data = zenniumConnectionLiveData.waitForBinaryTelegram()
            packetId = data[0]
            data = data[1:]
            """
            Type:
            1 = Init measurement begin
            2 = Measurement end
            4 = Measurement data names
            5 = Measurement data units
            6 = ASCII data
            """
            if packetId in [1, 2, 4, 5, 6]:
                print(data.decode("ASCII"))
        except:
            r"""
            The connection to the term has an error or the socket has been closed.
            """
            print("term error live thread")
            keepThreadRunning = False

    print("live thread left")
    return


if __name__ == "__main__":
    zenniumConnectionLiveData = ThalesRemoteConnection()
    zenniumConnectionLiveData.connectToTerm("localhost", "Logging")

    liveThread = threading.Thread(target=liveDataThreadFunction)
    liveThread.start()

zenniumConnection = ThalesRemoteConnection()
zenniumConnection.connectToTerm("localhost", "ScriptRemote")

zahnerZennium = ThalesRemoteScriptWrapper(zenniumConnection)
zahnerZennium.forceThalesIntoRemoteScript()
zahnerZennium.hideWindow()

zahnerZennium.calibrateOffsets()

watchThread = threading.Thread(target=watchThreadFunction)
watchThread.start()

zahnerZennium.setPotentiostatMode(PotentiostatMode.POTMODE_POTENTIOSTATIC)
zahnerZennium.setAmplitude(10e-3)
zahnerZennium.setPotential(0)
zahnerZennium.setLowerFrequencyLimit(750)
zahnerZennium.setStartFrequency(1000)
zahnerZennium.setUpperFrequencyLimit(1500)
zahnerZennium.setLowerNumberOfPeriods(2)
zahnerZennium.setLowerStepsPerDecade(2)
zahnerZennium.setUpperNumberOfPeriods(2)
zahnerZennium.setUpperStepsPerDecade(20)
zahnerZennium.setScanDirection("startToMax")
zahnerZennium.setScanStrategy("single")

zahnerZennium.enablePotentiostat()


zahnerZennium.setFrequency(1)
zahnerZennium.setAmplitude(10e-3)
zahnerZennium.setNumberOfPeriods(3)

print("measurement start")

zahnerZennium.measureEIS()
for i in range(20):
    zahnerZennium.getPotential()
    zahnerZennium.setPotential(0)

print("measurement end")

zahnerZennium.disablePotentiostat()

print("set thread kill flag")
keepThreadRunning = False

print("disconnect connections")
zahnerZennium.showWindow()
zenniumConnection.disconnectFromTerm()
zenniumConnectionLiveData.disconnectFromTerm()

print("join the threads")
liveThread.join()
watchThread.join()

print("finish")



r"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2024 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import socket
import time
import struct
import threading
import queue
from typing import Optional, Union
from _socket import SHUT_RD
from thales_remote.error import TermConnectionError

from datetime import datetime


class ThalesRemoteConnection(object):
    r"""
    Class to handle the Thales remote connection.
    """

    _term_port: int
    _socket_handle: Optional[socket.socket]
    _receiving_worker: Optional[threading.Thread]
    _send_mutex: threading.Semaphore
    _receiving_worker_is_running: bool
    _available_channels: list[int]
    _queuesForChannels: dict[int, queue.Queue[Optional[bytes]]]
    _connectionName: str

    def __init__(self):
        self._term_port = 260  # The port used by Thales
        self._socket_handle = None
        self._receiving_worker = None
        self._send_mutex = threading.Semaphore(1)
        self._receiving_worker_is_running = False
        self._available_channels = [2, 42, 128, 129, 130, 131, 132]
        self._queuesForChannels = dict()

        for channel in self._available_channels:
            self._queuesForChannels[channel] = queue.Queue()

        self._connectionName = ""

    # methods for context handler
    # documentation: https://docs.python.org/3/reference/datamodel.html#context-managers
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._socket_handle is not None:
            self.disconnectFromTerm()

    def connectToTerm(
        self, address: str, connection_name: str = "ScriptRemote"
    ) -> bool:
        r"""
        Connect to the Term Software with a IP address and an optional connection name.

        *connection_name* cannot be chosen completely freely.
        If you want to send Remote2 commands with the connection,
        then the name **ScriptRemote** must be used and if you want to receive the online display data, then the name **Logging** must be used.

        :param address: hostname or ip-address of the host running "Term" application
        :param connection_name: name of the connection. *ScriptRemote* for Remote2 and *Logging* as Online Display are fixed.
        :returns: True on success, False on failure
        """
        time.sleep(0.4)
        self._socket_handle = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self._socket_handle.connect((address, self._term_port))
        except:
            raise TermConnectionError("Connection to the Term not possible.")

        self._startTelegramListener()

        time.sleep(0.4)

        self._connectionName = connection_name
        payload_length = len(connection_name)

        registration_packet = bytearray()
        registration_packet += bytearray(struct.pack(">H", payload_length))
        registration_packet += bytearray([0x12, 0xD0, 0xFF, 0xFF, 0xFF, 0xFF])
        registration_packet += bytearray(connection_name, "ASCII")

        # print("\n" + str(datetime.now().time()) + " send:")
        # print(f"payload_length: {payload_length}")
        # print(f"registration_packet: {registration_packet}")
        # print("complete packet:" + str(registration_packet))

        self._send_mutex.acquire()
        self.sendall(registration_packet)
        self._send_mutex.release()

        time.sleep(0.8)

        return True

    def sendall(self, data: bytearray):
        return self._socket_handle.sendall(data)

    def readall(self, numBytes: int) -> bytes:
        bytesToRead = numBytes
        readBytes = bytes()
        while bytesToRead > 0:
            readBytes += self._socket_handle.recv(bytesToRead)
            bytesToRead = numBytes - len(readBytes)
        return readBytes

    def getConnectionName(self) -> str:
        r"""
        get the connection name

        :returns: name of the connection
        """
        return self._connectionName

    def disconnectFromTerm(self) -> None:
        r"""
        close the connection to Term and cleanup

        Stops the thread used for receiving telegrams assynchronously and shuts down
        the network connection. Put None into the Queues to free the waiting threads.
        They wait in waitForBinaryTelegram and if they receive None from the Queue, the will throw an exception.
        """
        self.sendStringAndWaitForReplyString(
            "3," + str(self._connectionName) + ",0,RS", 128
        )
        time.sleep(0.2)
        self.sendTelegram(bytearray([255, 255]), 4)
        time.sleep(0.2)
        self._stopTelegramListener()
        time.sleep(0.2)
        self._closeSocket()
        for key in self._queuesForChannels.keys():
            self._queuesForChannels[key].put(None)
        return

    def isConnectedToTerm(self) -> bool:
        r"""
        check if the connection to Term is open

        :returns: True if connected, False otherwise
        """
        return self._socket_handle != None

    def sendTelegram(
        self,
        payload: Union[str, bytearray],
        message_type: int,
        timeout: Optional[float] = None,
    ) -> None:
        r"""
        send a telegram (data) to Term

        Sending a telegram to the term.
        If the other thread hangs while sending and the semaphore cannot be aquired within the timout,
        then a TermConnectionError is thrown. If a timeout occurs in the socket,
        then an exception is thrown by the socket.

        :param payload: The actual data which is being sent to Term. This can be a string or an bytearray.
        :param message_type: Used internally by the DevCli dll. Depends on context. Most of the time 2.
        :param timeout: The timeout for sending data in seconds, blocking at None.
        """
        packet = bytearray()
        data = bytearray()

        if isinstance(payload, str):
            # datatype string
            payload_length = len(payload)
            data += bytearray(payload, "ASCII")
        else:
            # datatype bytearray
            payload_length = len(payload)
            data = payload

        packet += bytearray(struct.pack("<H", payload_length))
        packet += bytearray(struct.pack("<B", message_type))
        packet += data

        if self._send_mutex.acquire(True, timeout=timeout):
            self._socket_handle.settimeout(timeout)
            try:
                # print("\n" + str(datetime.now().time()) + " send:")
                # print(f"payload_length: {payload_length} message_type: {message_type}")
                # print(f"payload: {data}")
                # print("complete packet:" + str(packet))

                self.sendall(packet)
            finally:
                self._socket_handle.settimeout(None)
                self._send_mutex.release()
        else:
            """
            The semaphore to send could not be aquired within the timeout.
            There must be an error in the connection to the term.
            """
            raise TermConnectionError("Socket error during data transmission.")
        return

    def waitForBinaryTelegram(
        self, message_type: int = 2, timeout: Optional[float] = None
    ) -> bytes:
        r"""
        block infinitely until the next Telegram is arriving

        If some Telegram has already arrived it will just return the last one from the queue.

        :param message_type: Used internally by the DevCli dll. Depends on context. Most of the time 2.
        :param timeout: The timeout for sending data in seconds, blocking at None
        :returns: The response from the device or an empty bytearray if someting went wrong.
        :rtype: bytearray
        """
        retval = self._queuesForChannels[message_type].get(True, timeout=timeout)
        if retval is None:
            raise TermConnectionError("Socket error during data reception.")
        return retval

    def waitForStringTelegram(
        self, message_type: int = 2, timeout: Optional[float] = None
    ) -> str:
        r"""
        block infinitely until the next Telegram is arriving

        If some Telegram has already arrived it will just return the last one from the queue.

        :param message_type: Used internally by the DevCli dll. Depends on context. Most of the time 2.
        :param timeout: The timeout for sending data in seconds, blocking at None
        :returns: The last received telegram or an empty string if someting went wrong.
        :rtype: string
        """
        retval = self.waitForBinaryTelegram(message_type, timeout).decode("ASCII")
        return retval

    def sendStringAndWaitForReplyString(
        self,
        payload: Union[str, bytearray],
        message_type: int,
        timeout: Optional[float] = None,
        answer_message_type: int = None,
    ) -> str:
        r"""
        convenience function: send a telegram and wait for it's reply

        If a timeout or a socket error occurs an exception is thrown.


        :param payload: The actual data which is being sent to Term. This can be a string or an bytearray.
        :param message_type: Used internally by the DevCli dll. Depends on context. Most of the time 2.
        :param timeout: The timeout for sending data in seconds, blocking at None.
        :returns: The last received telegram or an empty string if someting went wrong.
        :rtype: string
        """
        if answer_message_type is None:
            answer_message_type = message_type
        self.sendTelegram(payload, message_type, timeout)
        return self.waitForStringTelegram(answer_message_type, timeout)

    # The following methods should not be called by the user.
    # They are marked with the prefix '_' after the Python convention for proteced.

    def _telegramListenerJob(self) -> None:
        r"""
        runs in a separate thread, pushing the incomming packets into the queues.
        """
        while self._receiving_worker_is_running:
            message_type, telegram = self._readTelegramFromSocket()
            if len(telegram) > 0 and message_type in self._available_channels:
                self._queuesForChannels[message_type].put(telegram)
            elif message_type is None:
                # An error has occurred in the connection. None is passed into all queues to free
                # the waiting threads from the queue. If they have received None, they throw an exception.
                # The thread is then exited.
                for key, value in self._queuesForChannels.items():
                    value.put(None)
                self._receiving_worker_is_running = False
        return

    def _startTelegramListener(self) -> None:
        r"""
        starts the thread handling the asyncronously incoming data
        """
        self._receiving_worker_is_running = True
        self._receiving_worker = threading.Thread(
            target=self._telegramListenerJob, daemon=True
        )
        self._receiving_worker.start()
        return

    def _stopTelegramListener(self) -> None:
        r"""
        stops the thread handling the incoming data gracefully
        """
        self._socket_handle.shutdown(SHUT_RD)
        self._receiving_worker_is_running = False
        self._receiving_worker.join()
        return

    def _readTelegramFromSocket(self) -> tuple[Optional[str], bytearray]:
        r"""
        reads the raw telegram structure from the socket stream

        When a socket exception occurs, None and an empty byte array are returned.
        The caller of the function then passes the None to the queue to raise an
        exception in the threads waiting at the queue.
        """
        try:
            header_len: bytes = self.readall(2)
            header_type_bytes: bytes = self.readall(1)
            header_type: str = struct.unpack("<B", header_type_bytes)[
                0
            ]  # actually a character, not a str
            incoming_packet: int = self.readall(struct.unpack("<H", header_len)[0])

            # print("\n" + str(datetime.now().time()) + " read:")
            # print(f"payload_length: {struct.unpack('<H', header_len)[0]} message_type: {header_type}")
            # print(f"payload: {incoming_packet}")
            # print("complete packet:" + str(header_len + header_type_bytes + incoming_packet))

        except:
            header_type = None
            incoming_packet = bytearray()
        return header_type, incoming_packet

    def _closeSocket(self) -> None:
        r"""
        close the socket
        """
        self._socket_handle.close()
        self._socket_handle = None
        return



import sys
from thales_remote.connection import ThalesRemoteConnection
from thales_remote.script_wrapper import (
    PotentiostatMode,
    ScanStrategy,
    ScanDirection,
    FileNaming,
    Pad4Mode,
    ThalesRemoteScriptWrapper,
)
from thales_remote.file_interface import ThalesFileInterface
from zahner_analysis.file_import.ism_import import IsmImport

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import EngFormatter
import os
import shutil

if __name__ == "__main__":
    remoteIP = "192.168.2.66"

    zenniumConnection = ThalesRemoteConnection()
    zenniumConnection.connectToTerm(remoteIP, "ScriptRemote")

    zahnerZennium = ThalesRemoteScriptWrapper(zenniumConnection)
    zahnerZennium.forceThalesIntoRemoteScript()

    zahnerZennium.calibrateOffsets()

fileInterface = ThalesFileInterface(remoteIP)

zahnerZennium.setEISOutputPath(r"C:\THALES\temp")
zahnerZennium.setEISNaming(FileNaming.INDIVIDUAL)
zahnerZennium.setEISOutputFileName("myeis")

zahnerZennium.setPotentiostatMode(PotentiostatMode.POTMODE_POTENTIOSTATIC)
zahnerZennium.setAmplitude(100e-3)
zahnerZennium.setPotential(0)
zahnerZennium.setLowerFrequencyLimit(10)
zahnerZennium.setStartFrequency(1000)
zahnerZennium.setUpperFrequencyLimit(10000)
zahnerZennium.setLowerNumberOfPeriods(5)
zahnerZennium.setLowerStepsPerDecade(5)
zahnerZennium.setUpperNumberOfPeriods(20)
zahnerZennium.setUpperStepsPerDecade(5)
zahnerZennium.setScanDirection(ScanDirection.START_TO_MAX)
zahnerZennium.setScanStrategy(ScanStrategy.SINGLE_SINE)

zahnerZennium.enablePotentiostat()

zahnerZennium.measureEIS()

zahnerZennium.disablePotentiostat()
zahnerZennium.setAmplitude(0)

file = fileInterface.acquireFile(r"C:\THALES\temp\myeis.ism")

fileHandle = open(r"C:\THALES\temp\manual_save.ism", "wb")
fileHandle.write(file.binaryData)
fileHandle.close()

ismFile = IsmImport(file.binaryData)

impedanceFrequencies = ismFile.getFrequencyArray()

impedanceAbsolute = ismFile.getImpedanceArray()
impedancePhase = ismFile.getPhaseArray()

figBode, (impedanceAxis, phaseAxis) = plt.subplots(2, 1, sharex=True)
figBode.suptitle("Bode")

impedanceAxis.loglog(impedanceFrequencies, impedanceAbsolute, marker="+", markersize=5)
impedanceAxis.xaxis.set_major_formatter(EngFormatter(unit="Hz"))
impedanceAxis.yaxis.set_major_formatter(EngFormatter(unit="$\Omega$"))
impedanceAxis.set_xlabel(r"$f$")
impedanceAxis.set_ylabel(r"$|Z|$")
impedanceAxis.grid(which="both")

phaseAxis.semilogx(
    impedanceFrequencies,
    np.abs(impedancePhase * (360 / (2 * np.pi))),
    marker="+",
    markersize=5,
)
phaseAxis.xaxis.set_major_formatter(EngFormatter(unit="Hz"))
phaseAxis.yaxis.set_major_formatter(EngFormatter(unit="$°$", sep=""))
phaseAxis.set_xlabel(r"$f$")
phaseAxis.set_ylabel(r"$|Phase|$")
phaseAxis.grid(which="both")
phaseAxis.set_ylim([0, 90])
figBode.set_size_inches(18, 12)
plt.show()

localDirectory = r"C:\THALES\temp\test"

# Delete the entire contents of the directory.
for file in os.listdir(localDirectory):
    fileWithPath = os.path.join(localDirectory, file)
    try:
        if os.path.isfile(fileWithPath) or os.path.islink(fileWithPath):
            os.unlink(fileWithPath)
        elif os.path.isdir(fileWithPath):
            shutil.rmtree(fileWithPath)
    except:
        pass

fileInterface.enableSaveReceivedFilesToDisk(path=localDirectory)
fileInterface.enableKeepReceivedFilesInObject()
fileInterface.enableAutomaticFileExchange()

zahnerZennium.setEISNaming("counter")
zahnerZennium.setEISCounter(13)
zahnerZennium.setEISOutputFileName("spectra")

zahnerZennium.setupPad4Channel(1, 1, 1, voltageRange=4.0, shuntResistor=10e-3)
zahnerZennium.setupPad4Channel(1, 2, 1, voltageRange=4.0, shuntResistor=10e-3)
zahnerZennium.setupPad4ModeGlobal(Pad4Mode.VOLTAGE)  # or Pad4Mode.CURRENT
zahnerZennium.enablePad4Global()

zahnerZennium.measureEIS()
fileInterface.disableAutomaticFileExchange()

zahnerZennium.measureEIS()

fileInterface.enableAutomaticFileExchange()
zahnerZennium.measureEIS()
zahnerZennium.setAmplitude(0)

for file in fileInterface.getReceivedFiles():
    ismFile = IsmImport(file.binaryData)
    print(f"{file.name} measurement finished at {ismFile.getMeasurementEndDateTime()}")

for file in os.listdir(localDirectory):
    print(file)

zenniumConnection.disconnectFromTerm()
fileInterface.close()



import sys
import math
import cmath
from thales_remote.connection import ThalesRemoteConnection
from thales_remote.script_wrapper import PotentiostatMode, ThalesRemoteScriptWrapper


def printImpedance(impedance):
    print(
        f"Impedance: {abs(impedance):>10.3e} ohm {cmath.phase(impedance)/cmath.pi*180.0:>10.2f} degree"
    )
    return


def spectrum(scriptHandle, lower_frequency, upper_frequency, number_of_points):
    log_lower_frequency = math.log(lower_frequency)
    log_upper_frequency = math.log(upper_frequency)
    log_interval_spacing = (log_upper_frequency - log_lower_frequency) / (
        number_of_points - 1
    )

    for i in range(number_of_points):
        current_frequency = math.exp(log_lower_frequency + log_interval_spacing * i)
        print(f"Frequency: {current_frequency:e} Hz")
        printImpedance(scriptHandle.getImpedance(current_frequency))

    return


TARGET_HOST = "localhost"

zenniumConnection = ThalesRemoteConnection()
zenniumConnection.connectToTerm(TARGET_HOST)

zahnerZennium = ThalesRemoteScriptWrapper(zenniumConnection)
zahnerZennium.forceThalesIntoRemoteScript()

zahnerZennium.calibrateOffsets()

zahnerZennium.setPotentiostatMode(PotentiostatMode.POTMODE_POTENTIOSTATIC)
zahnerZennium.setPotential(1.0)
zahnerZennium.enablePotentiostat()

for i in range(5):
    print(f"Potential:\t{zahnerZennium.getPotential():>10.6f} V")
    print(f"Current:\t{zahnerZennium.getCurrent():>10.3e} A")

zahnerZennium.disablePotentiostat()
zahnerZennium.setPotentiostatMode(PotentiostatMode.POTMODE_GALVANOSTATIC)
zahnerZennium.setCurrent(20e-9)
zahnerZennium.enablePotentiostat()

for i in range(5):
    print(f"Potential:\t{zahnerZennium.getPotential():>10.6f} V")
    print(f"Current:\t{zahnerZennium.getCurrent():>10.3e} A")

zahnerZennium.disablePotentiostat()
zahnerZennium.setPotentiostatMode(PotentiostatMode.POTMODE_POTENTIOSTATIC)
zahnerZennium.setPotential(1.0)
zahnerZennium.enablePotentiostat()
zahnerZennium.setFrequency(2000)
zahnerZennium.setNumberOfPeriods(3)

zahnerZennium.enablePotentiostat()
zahnerZennium.getCurrent()

zahnerZennium.setAmplitude(10e-3)

printImpedance(zahnerZennium.getImpedance())
printImpedance(zahnerZennium.getImpedance(2000))
printImpedance(zahnerZennium.getImpedance(2000, 10e-3, 3))

spectrum(zahnerZennium, 1000, 2e5, 10)

zahnerZennium.disablePotentiostat()
zahnerZennium.setAmplitude(0)

zenniumConnection.disconnectFromTerm()
print("finish")



r"""
  ____       __                        __    __   __      _ __
 /_  / ___ _/ /  ___  ___ ___________ / /__ / /__/ /_____(_) /__
  / /_/ _ `/ _ \/ _ \/ -_) __/___/ -_) / -_)  '_/ __/ __/ /  '_/
 /___/\_,_/_//_/_//_/\__/_/      \__/_/\__/_/\_\\__/_/ /_/_/\_\

Copyright 2024 Zahner-Elektrik GmbH & Co. KG

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
"""

import os
import threading
import time
from queue import Empty
from typing import Optional, Union, ByteString

from thales_remote.connection import ThalesRemoteConnection
from dataclasses import dataclass
import warnings


@dataclass
class FileData:
    name: str
    path: str
    binaryData: ByteString


class ThalesFileInterface(object):
    r"""Class which realizes the file transfer between Term software and Python.

    The measurement files cannot be stored on network drives with the Term software, therefore
    this class was implemented. If Python runs on another computer and controls Thales via network,
    then with this class the measurement result files can be exchanged via network.

    This class establishes an additional socket connection to the Term, which can be used to transfer
    individual files manually. Or you can set that all ism, isc or isw files are transferred automatically
    at the end of the measurement.

    :param address: IP address of the computer running the Term software.
    :param connectionName: The name of the connection default FileExchange. But can also be freely assigned.
    """

    _device_name: str
    remoteConnection: ThalesRemoteConnection
    _receiver_is_running: bool
    _automatic_file_exchange: bool
    _files_to_skip: list[str]
    receivedFiles: list[FileData]
    pathToSave: str
    _save_received_files_to_disk: bool
    _keep_files_in_object: bool

    def __init__(self, address, connectionName="FileExchange"):
        self._device_name = connectionName
        self.remoteConnection = ThalesRemoteConnection()
        self.remoteConnection.connectToTerm(address, self._device_name)
        self._receiver_is_running = False
        self._automatic_file_exchange = False
        self._files_to_skip = ["lastshot.ism"]
        self.receivedFiles = []
        self.pathToSave = os.getcwd()
        self._save_received_files_to_disk = False
        self._keep_files_in_object = True
        return

    def close(self) -> None:
        r"""
        close the file interface

        The automatic file sending is disabled and the socket connection is closed.
        """
        self.disableAutomaticFileExchange()
        self.remoteConnection.disconnectFromTerm()
        return

    def enableAutomaticFileExchange(
        self, enable: bool = True, fileExtensions: str = "*.ism*.isc*.isw*.iss"
    ) -> str:
        r"""
        turn on automatic file exchange

        If automatic file exchange is enabled, the files configured with the fileExtensions variable are transferred automatically.

        In order for the files to be automatically saved to disk, this must be configured with the :func:`~thales_remote.file_interface.ThalesFileInterface.enableSaveReceivedFilesToDisk` method.
        With the method :func:`~thales_remote.file_interface.ThalesFileInterface.enableKeepReceivedFilesInObject`
        can be set whether the files should still remain in the Python object, so that you can read
        the received files as an array with :func:`~thales_remote.file_interface.ThalesFileInterface.getReceivedFiles`.
        The standard setting is that the files remain in the object.

        :param enable: Enable automatic file exchange. Default = True.
        :param fileExtensions: File extensions that will be exchanged. You can see the default paramter.
            But you can also specify only one type of file or more.
        :type fileExtensions: string
        :returns: The response string from the device.
        :rtype: string
        """
        if enable:
            retval = self.remoteConnection.sendStringAndWaitForReplyString(
                f"3,{self._device_name},4,ON,{fileExtensions}",
                message_type=128,
                answer_message_type=132,
            )
            self._startWorker()
        else:
            # Sending the command that no more data should be sent.
            # Then wait 1 second until the Thales has possibly sent everything.
            # Then stopp the worker thread.
            retval = self.remoteConnection.sendStringAndWaitForReplyString(
                f"3,{self._device_name},4,OFF",
                message_type=128,
                answer_message_type=132,
            )
            time.sleep(1)
            self._stopWorker()
        return retval

    def appendFilesToSkip(self, file: Union[str, list[str]]):
        r"""
        set filenames to be filtered and not processed by Python

        Files with these names are not saved to disk by Python and do not remain in the object.

        :param file: Filename or list with filenames to be filtered.
        """
        if isinstance(file, list):
            self._files_to_skip.extend(file)
        else:
            self._files_to_skip.append(file)

    def disableAutomaticFileExchange(self) -> str:
        r"""
        turn off automatic file exchange

        :returns: the response string from the device
        """
        return self.enableAutomaticFileExchange(False)

    def aquireFile(self, filename: str) -> Union[FileData, None]:
        r"""
        transfer a single file, deprecated because misspelled.

        This command transfers a single from Term to Python.
        **This command can only be executed if automatic transfer is disabled.** If automatic transfer is enabled, None is returned.
        The parameter filename is used to specify the full path of the file, on the computer running
        the Thales software, to be transferred e.g. r"C:\\THALES\\temp\\test1\\myeis.ism".

        The function returns the file as dataclass.

        :param filename: Filename with path on the Thales computer.
        :returns: A dataclass with the file or None if this command is called when automatic file
            exchange is activated.
        """
        warnings.warn(
            "This function is misspelled and will be removed, use acquireFile instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.acquireFile(filename)

    def acquireFile(self, filename: str) -> Union[FileData, None]:
        r"""
        transfer a single file

        This command transfers a single from Term to Python.
        **This command can only be executed if automatic transfer is disabled.** If automatic transfer is enabled, None is returned.
        The parameter filename is used to specify the full path of the file, on the computer running
        the Thales software, to be transferred e.g. r"C:\\THALES\\temp\\test1\\myeis.ism".

        The function returns the file as dataclass.

        :param filename: Filename with path on the Thales computer.
        :returns: A dataclass with the file or None if this command is called when automatic file
            exchange is activated.
        """
        if self._receiver_is_running:
            return None
        else:
            self.remoteConnection.sendTelegram(
                f"3,{self._device_name},1,{filename}", message_type=128
            )
            return self._receiveFile()

    def setSavePath(self, path: str) -> None:
        r"""
        set the path where the files should be saved on the local computer

        This command sets only the path. If the path does not exist, it will be created.
        The path must be accessible by Python, otherwise there are no restrictions on the path.

        :param path: Path where the files should be saved. For example r"D:\\myLocalDirectory".
        """
        self.pathToSave = path
        os.makedirs(self.pathToSave, exist_ok=True)
        return

    def enableSaveReceivedFilesToDisk(
        self, enable: bool = True, path: Optional[str] = None
    ) -> None:
        r"""
        enable the automatic saving of files to the hard disk

        This command configures that the files are automatically saved by Python.
        With this command the path can be passed with the optional parameter path.
        The path must be accessible by Python, otherwise there are no restrictions on the path.
        Default path is the current working directory of Python.

        :param enable: Enable automatic file saving to the hard disk. Default = True.
        :param path: Optional path where the files should be saved. For example r"D:\\myLocalDirectory".
        """
        if path is not None:
            self.setSavePath(path)
        self._save_received_files_to_disk = enable
        return

    def disableSaveReceivedFilesToDisk(self) -> None:
        r"""
        disable the automatic saving of files to the hard disk
        """
        return self.enableSaveReceivedFilesToDisk(False)

    def enableKeepReceivedFilesInObject(self, enable: bool = True) -> None:
        r"""Enable that the files remain in the Python object.

        If you perform many measurements, the Python object would grow larger and larger due to the
        number of files, so you can disable that the files are stored in the object as an array.
        By default, the files remain in the object.

        :param enable: True = Allow the files to remain in the object.
        """
        self._keep_files_in_object = enable
        return

    def disableKeepReceivedFilesInObject(self) -> None:
        r"""
        disable that the files remain in the Python object
        """
        return self.enableKeepReceivedFilesInObject(False)

    def getReceivedFiles(self) -> list[FileData]:
        r"""
        read all files from the Python object

        This function returns an array. Each element in the array is a dictionary as described in
        function :func:`~thales_remote.file_interface.ThalesFileInterface.aquireFile`.

        :returns: Array with the files from the Python object.
        """
        return self.receivedFiles

    def getLatestReceivedFile(self) -> FileData:
        r"""
        read the latest files from the Python object

        :returns: latest file from the Python object
        """
        return self.receivedFiles[-1]

    def deleteReceivedFiles(self) -> None:
        r"""
        delete all files from the Python object
        """
        self.receivedFiles = []
        return

    # The following methods should not be called by the user.
    # They are marked with the prefix '_' after the Python convention for proteced.

    def _saveReceivedFile(self, fileToWrite: FileData) -> None:
        r"""
        saves the passed file to disk.
        """
        fileNameWithPath = os.path.join(self.pathToSave, fileToWrite.name)

        with open(fileNameWithPath, "wb") as file:
            file.write(fileToWrite.binaryData)
        return

    def _receiveFile(self, timeout: Optional[float] = None) -> Union[FileData, None]:
        r"""
        receive one file with optional timeout

        :param timeout: receive timeout
        :returns: structure with the file
        """

        try:
            filePath: str = self.remoteConnection.waitForStringTelegram(
                130, timeout=timeout
            )
        except Empty:
            return None

        fileLength = int(self.remoteConnection.waitForStringTelegram(129))
        bytesToReceive = fileLength

        fileData = bytearray()
        while bytesToReceive > 0:
            receivedBytes = self.remoteConnection.waitForBinaryTelegram(131)
            fileData += receivedBytes
            bytesToReceive -= len(receivedBytes)

        fileSplit = filePath.split("\\")
        fileName = fileSplit[-1]

        return FileData(fileName, filePath, fileData)

    def _startWorker(self) -> None:
        r"""
        starts the thread handling the asyncronously incoming data
        """
        if not self._receiver_is_running:
            self._receiver_is_running = True
            self.receivingWorker = threading.Thread(target=self._fileReceiverJob)
            self.receivingWorker.start()
        return

    def _stopWorker(self) -> None:
        r"""
        stops the thread handling the incoming data gracefully
        """
        if self._receiver_is_running:
            self._receiver_is_running = False
            self.receivingWorker.join()
        return

    def _fileReceiverJob(self) -> None:
        r"""
        method running in a separate thread; manages the received files
        """
        while self._receiver_is_running:
            try:
                file = self._receiveFile(1)
                if file is not None:
                    if file.name not in self._files_to_skip:
                        if self._keep_files_in_object:
                            self.receivedFiles.append(file)

                        if self._save_received_files_to_disk:
                            self._saveReceivedFile(file)
            except:
                self._receiver_is_running = False
        return



from thales_remote.connection import ThalesRemoteConnection
from thales_remote.script_wrapper import (
    PotentiostatMode,
    ScanStrategy,
    ScanDirection,
    FileNaming,
    Pad4Mode,
    ThalesRemoteScriptWrapper,
)

from zahner_analysis.file_import.ism_import import IsmImport
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.ticker import EngFormatter

zenniumConnection = ThalesRemoteConnection()
zenniumConnection.connectToTerm("192.168.2.47", "ScriptRemote")

zahnerZennium = ThalesRemoteScriptWrapper(zenniumConnection)
zahnerZennium.forceThalesIntoRemoteScript()

zahnerZennium.calibrateOffsets()

zahnerZennium.setEISNaming(FileNaming.COUNTER)
zahnerZennium.setEISCounter(1)
zahnerZennium.setEISOutputPath(r"C:\THALES\temp\test1")
zahnerZennium.setEISOutputFileName("spectra_cells")

zahnerZennium.setPotentiostatMode(PotentiostatMode.POTMODE_POTENTIOSTATIC)
zahnerZennium.setAmplitude(100e-3)
zahnerZennium.setPotential(0)
zahnerZennium.setLowerFrequencyLimit(0.1)
zahnerZennium.setStartFrequency(1000)
zahnerZennium.setUpperFrequencyLimit(10000)
zahnerZennium.setLowerNumberOfPeriods(2)
zahnerZennium.setLowerStepsPerDecade(5)
zahnerZennium.setUpperNumberOfPeriods(20)
zahnerZennium.setUpperStepsPerDecade(10)
zahnerZennium.setScanDirection(ScanDirection.START_TO_MAX)
zahnerZennium.setScanStrategy(ScanStrategy.SINGLE_SINE)

zahnerZennium.setPotential(1)
zahnerZennium.enablePotentiostat()

print(f"ACQ-Setup: {zahnerZennium.readAcqSetup()}")

for i in range(2):
    print(f"Current:\t{zahnerZennium.getCurrent():>10.3e} A")
    print(f"Potential:\t{zahnerZennium.getPotential():>10.6f} V")
    print(zahnerZennium.readAllAcqChannels())
    print(f"PAD4 Channel: {0} = {zahnerZennium.readAcqChannel(0)}")
    print(f"PAD4 Channel: {1} = {zahnerZennium.readAcqChannel(1)}")


zahnerZennium.setPotential(0)

zahnerZennium.setupPad4Channel(1, 1, 1, voltageRange=4.0, shuntResistor=10e-3)
zahnerZennium.setupPad4Channel(1, 2, 1, voltageRange=4.0, shuntResistor=10e-3)
zahnerZennium.setupPad4Channel(1, 3, 0)
zahnerZennium.setupPad4Channel(1, 4, 0)
zahnerZennium.setupPad4ModeGlobal(Pad4Mode.VOLTAGE)  # or Pad4Mode.CURRENT
zahnerZennium.enablePad4Global()

for frequency in [100, 300, 1000]:
    print(f"frequency: {frequency}")
    impedanceResult = zahnerZennium.getImpedancePad4(frequency=frequency)

    for key, val in impedanceResult.items():
        print(f"{key}: {val}")

zahnerZennium.enablePotentiostat()
zahnerZennium.measureEIS()
zahnerZennium.disablePotentiostat()
zahnerZennium.setAmplitude(0)
zenniumConnection.disconnectFromTerm()

ismFileStack = IsmImport(r"C:\THALES\temp\test1\spectra_cells_0001_ser00.ism")
impedanceFrequenciesStack = ismFileStack.getFrequencyArray()
impedanceAbsoluteStack = ismFileStack.getImpedanceArray()
impedancePhaseStack = ismFileStack.getPhaseArray()
impedanceComplexStack = ismFileStack.getComplexImpedanceArray()

ismFileCell1 = IsmImport(r"C:\THALES\temp\test1\spectra_cells_0001_ser01.ism")
impedanceFrequenciesCell1 = ismFileCell1.getFrequencyArray()
impedanceAbsoluteCell1 = ismFileCell1.getImpedanceArray()
impedancePhaseCell1 = ismFileCell1.getPhaseArray()
impedanceComplexCell1 = ismFileCell1.getComplexImpedanceArray()

ismFileCell2 = IsmImport(r"C:\THALES\temp\test1\spectra_cells_0001_ser02.ism")
impedanceFrequenciesCell2 = ismFileCell2.getFrequencyArray()
impedanceAbsoluteCell2 = ismFileCell2.getImpedanceArray()
impedancePhaseCell2 = ismFileCell2.getPhaseArray()
impedanceComplexCell2 = ismFileCell2.getComplexImpedanceArray()

figNyquist, (nyquistAxis) = plt.subplots(1, 1)
figNyquist.suptitle("Nyquist")

nyquistAxis.plot(
    np.real(impedanceComplexStack),
    -np.imag(impedanceComplexStack),
    marker="x",
    markersize=5,
    label="Stack",
)
nyquistAxis.plot(
    np.real(impedanceComplexCell1),
    -np.imag(impedanceComplexCell1),
    marker="x",
    markersize=5,
    label="Cell 1",
)
nyquistAxis.plot(
    np.real(impedanceComplexCell2),
    -np.imag(impedanceComplexCell2),
    marker="x",
    markersize=5,
    label="Cell 2",
)

nyquistAxis.grid(which="both")
nyquistAxis.set_aspect("equal")
nyquistAxis.xaxis.set_major_formatter(EngFormatter(unit=r"$\Omega$"))
nyquistAxis.yaxis.set_major_formatter(EngFormatter(unit=r"$\Omega$"))
nyquistAxis.set_xlabel(r"$Z_{\rm re}$")
nyquistAxis.set_ylabel(r"$-Z_{\rm im}$")
nyquistAxis.legend(fontsize="large")
figNyquist.set_size_inches(20, 8)
plt.show()
figNyquist.savefig("nyquist.svg")



from thales_remote.epc_scpi_handler import EpcScpiHandlerFactory, EpcScpiHandler
from thales_remote.script_wrapper import PotentiostatMode
from zahner_potentiostat.scpi_control.datahandler import DataManager
from zahner_potentiostat.display.onlinedisplay import OnlineDisplay
from zahner_potentiostat.scpi_control.datareceiver import TrackTypes

import threading
from datetime import datetime


def getFileName(channel, cycle):
    time = str(datetime.now().time())
    time = time.replace(":", "")
    time = time.replace(",", "")
    time = time.replace(".", "")
    return f"channel{channel}_cycle{cycle}"  # + time


def channel1Thread(deviceHandler, channel=0):
    deviceHandler.scpiInterface.setMaximumTimeParameter(15)
    deviceHandler.scpiInterface.setParameterLimitCheckToleranceTime(0.1)

    configuration = {
        "figureTitle": "Online Display Channel 1",
        "xAxisLabel": "Time",
        "xAxisUnit": "s",
        "xTrackName": TrackTypes.TIME.toString(),
        "yAxis": [
            {
                "label": "Voltage",
                "unit": "V",
                "trackName": TrackTypes.VOLTAGE.toString(),
            },
            {
                "label": "Current",
                "unit": "A",
                "trackName": TrackTypes.CURRENT.toString(),
            },
        ],
    }

    for i in range(3):
        filename = getFileName(channel=channel, cycle=i)

        onlineDisplay = OnlineDisplay(
            deviceHandler.scpiInterface.getDataReceiver(),
            displayConfiguration=configuration,
        )

        deviceHandler.scpiInterface.measureOCVScan()

        for n in range(2):
            deviceHandler.scpiInterface.measureCharge(
                current=1e-3, stopVoltage=2, maximumTime="5 min"
            )

            deviceHandler.scpiInterface.measureDischarge(
                current=-1e-3, stopVoltage=0.5, maximumTime="5 min"
            )

        deviceHandler.scpiInterface.setMaximumTimeParameter(15)
        while deviceHandler.acquireSharedZennium(blocking=False) == False:
            deviceHandler.scpiInterface.measureOCVScan()

        dataManager = DataManager(deviceHandler.scpiInterface.getDataReceiver())
        dataManager.saveDataAsText(filename + ".txt")

        onlineDisplay.close()
        del onlineDisplay

        deviceHandler.switchToEPC()

        deviceHandler.sharedZenniumInterface.setEISNaming("individual")
        deviceHandler.sharedZenniumInterface.setEISOutputPath(
            r"C:\THALES\temp\multichannel"
        )
        deviceHandler.sharedZenniumInterface.setEISOutputFileName(filename)

        deviceHandler.sharedZenniumInterface.setPotentiostatMode(
            PotentiostatMode.POTMODE_POTENTIOSTATIC
        )
        deviceHandler.sharedZenniumInterface.setAmplitude(10e-3)
        deviceHandler.sharedZenniumInterface.setPotential(0)
        deviceHandler.sharedZenniumInterface.setLowerFrequencyLimit(100)
        deviceHandler.sharedZenniumInterface.setStartFrequency(500)
        deviceHandler.sharedZenniumInterface.setUpperFrequencyLimit(1000)
        deviceHandler.sharedZenniumInterface.setLowerNumberOfPeriods(5)
        deviceHandler.sharedZenniumInterface.setLowerStepsPerDecade(2)
        deviceHandler.sharedZenniumInterface.setUpperNumberOfPeriods(20)
        deviceHandler.sharedZenniumInterface.setUpperStepsPerDecade(5)
        deviceHandler.sharedZenniumInterface.setScanDirection("startToMax")
        deviceHandler.sharedZenniumInterface.setScanStrategy("single")

        deviceHandler.sharedZenniumInterface.measureEIS()

        deviceHandler.switchToSCPIAndReleaseSharedZennium()

    return


def channel2Thread(deviceHandler):
    deviceHandler.scpiInterface.setMaximumTimeParameter(15)
    deviceHandler.scpiInterface.setParameterLimitCheckToleranceTime(0.1)

    configuration = {
        "figureTitle": "Online Display Channel 2",
        "xAxisLabel": "Time",
        "xAxisUnit": "s",
        "xTrackName": TrackTypes.TIME.toString(),
        "yAxis": [
            {
                "label": "Voltage",
                "unit": "V",
                "trackName": TrackTypes.VOLTAGE.toString(),
            },
            {
                "label": "Current",
                "unit": "A",
                "trackName": TrackTypes.CURRENT.toString(),
            },
        ],
    }

    for i in range(2):
        filename = getFileName(channel=2, cycle=i)

        onlineDisplay = OnlineDisplay(
            deviceHandler.scpiInterface.getDataReceiver(),
            displayConfiguration=configuration,
        )

        deviceHandler.scpiInterface.measureOCVScan()

        for n in range(2):
            deviceHandler.scpiInterface.measureCharge(
                current=4, stopVoltage=1, maximumTime="5 min"
            )

            deviceHandler.scpiInterface.measureDischarge(
                current=-4, stopVoltage=0.6, maximumTime="5 min"
            )

        dataManager = DataManager(deviceHandler.scpiInterface.getDataReceiver())
        dataManager.saveDataAsText(filename + ".txt")

        onlineDisplay.close()
        del onlineDisplay

        deviceHandler.acquireSharedZennium()
        deviceHandler.switchToEPC()

        deviceHandler.sharedZenniumInterface.setEISNaming("individual")
        deviceHandler.sharedZenniumInterface.setEISOutputPath(
            r"C:\THALES\temp\multichannel"
        )
        deviceHandler.sharedZenniumInterface.setEISOutputFileName(filename)

        deviceHandler.sharedZenniumInterface.setPotentiostatMode(
            PotentiostatMode.POTMODE_POTENTIOSTATIC
        )
        deviceHandler.sharedZenniumInterface.setAmplitude(10e-3)
        deviceHandler.sharedZenniumInterface.setPotential(0)
        deviceHandler.sharedZenniumInterface.setLowerFrequencyLimit(100)
        deviceHandler.sharedZenniumInterface.setStartFrequency(500)
        deviceHandler.sharedZenniumInterface.setUpperFrequencyLimit(1000)
        deviceHandler.sharedZenniumInterface.setLowerNumberOfPeriods(5)
        deviceHandler.sharedZenniumInterface.setLowerStepsPerDecade(2)
        deviceHandler.sharedZenniumInterface.setUpperNumberOfPeriods(20)
        deviceHandler.sharedZenniumInterface.setUpperStepsPerDecade(5)
        deviceHandler.sharedZenniumInterface.setScanDirection("startToMax")
        deviceHandler.sharedZenniumInterface.setScanStrategy("single")

        deviceHandler.sharedZenniumInterface.measureEIS()

        deviceHandler.switchToSCPIAndReleaseSharedZennium()

    return


if __name__ == "__main__":
    handlerFactory = EpcScpiHandlerFactory()

    XPOT2 = handlerFactory.createEpcScpiHandler(epcChannel=1, serialNumber=27000)
    PP242 = handlerFactory.createEpcScpiHandler(epcChannel=4, serialNumber=35000)

    channel1ThreadHandler = threading.Thread(target=channel1Thread, args=(XPOT2, 1))
    channel2ThreadHandler = threading.Thread(target=channel2Thread, args=(PP242,))

    channel1ThreadHandler.start()
    channel2ThreadHandler.start()

    channel1ThreadHandler.join()
    channel2ThreadHandler.join()

    handlerFactory.closeAll()
    print("finish")



import sys
from thales_remote.connection import ThalesRemoteConnection
from thales_remote.script_wrapper import ThalesRemoteScriptWrapper

zenniumConnection = ThalesRemoteConnection()
zenniumConnection.connectToTerm("localhost", "ScriptRemote")

zahnerZennium = ThalesRemoteScriptWrapper(zenniumConnection)
zahnerZennium.forceThalesIntoRemoteScript()

zahnerZennium.calibrateOffsets()
zahnerZennium.setSequenceNaming("dateTime")
zahnerZennium.setSequenceOutputPath(r"C:\THALES\temp\test1")
zahnerZennium.setSequenceOutputFileName("batterysequence")

zahnerZennium.selectSequence(0)

zahnerZennium.enableSequenceAcqGlobal()
zahnerZennium.enableSequenceAcqChannel(3)

# disable
# zahnerZennium.disableSequenceAcqChannel(1)

zahnerZennium.runSequence()

zahnerZennium.selectPotentiostat(1)

zahnerZennium.setSequenceNaming("counter")
zahnerZennium.setSequenceCounter(13)
zahnerZennium.setSequenceOutputPath(r"C:\THALES\temp\test1")
zahnerZennium.setSequenceOutputFileName("batterysequence")

zahnerZennium.runSequenceFile(r"C:\Users\XXX\Desktop\myZahnerSequence.seq")

zahnerZennium.selectPotentiostat(0)

zenniumConnection.disconnectFromTerm()
print("finish")
